from functools import partial

import beartype.typing as tp
import jax.numpy as jnp
import jax.random as jr
import tensorflow_probability.substrates.jax.bijectors as tfb
import tensorflow_probability.substrates.jax.distributions as tfd
from flax import nnx
from flax.core import FrozenDict
from jax import jit, lax, vmap
from jaxtyping import Bool, Float, Int, PRNGKeyArray

from gallifrey.kernels.atoms import AbstractAtom, AbstractOperator
from gallifrey.kernels.library import KernelLibrary
from gallifrey.kernels.tree import TreeKernel
from gallifrey.utils.tree_helper import (
    calculate_max_depth,
    calculate_max_leaves,
    calculate_max_nodes,
    calculate_max_stack_size,
    get_child_idx,
    get_depth,
)
from gallifrey.utils.typing import ScalarBool, ScalarFloat, ScalarInt


class KernelPrior:
    """
    A prior distribution over kernel structures and parameters.

    Attributes
    ----------
    kernel_library : KernelLibrary
        An instance of the KernelLibrary class, containing
        the kernel classes and operators, and transformation functions.
    kernel_structure_prior : TreeStructurePrior
        A prior distribution over kernel structures.
    parameter_prior : ParameterPrior
        A prior distribution over kernel parameters.
    max_depth : ScalarInt
        The maximum depth of the nested kernel structure tree.
    num_datapoints : ScalarInt
        The number of data points in the dataset.
    max_kernel_parameter : int
        The maximum number of kernel parameters (max_atom_parameters
        * max_leaves).

    graphdef : nnx.GraphDef
        The graph definition of the kernel. Used together with the
        kernel state to create a TreeKernel object.

    kernels : tp.List[tp.Type[TreeKernel]]
        A list of kernel classes. (inherited from the kernel_library,
        see gallifrey.kernels.library.KernelLibrary)
    operators : tp.List[tp.Callable]
        A list of operators. (inherited from the kernel_library,
        see gallifrey.kernels.library.KernelLibrary)
    is_operator : Bool[jnp.ndarray, " D"]
        An array that indicates whether each kernel in the library is an operator.
        (inherited from the kernel_library, see gallifrey.kernels.library.KernelLibrary)
    probs :  Float[jnp.ndarray, " D"]
        The probabilities of sampling each kernel (or operator) in the library.
        (inherited from the kernel_structure_prior,
          see gallifrey.kernels.prior.TreeStructurePrior)
    prior_transforms : FrozenDict[str, tfb.Bijector]
        A (frozen) dictionary containing bijectors for transforming the
        distribution of the sampled parameters from a standard
        normal to the desired prior distribution. (inherited from the kernel_library,
        see gallifrey.kernels.library.KernelLibrary)
    support_bijectors : tuple[tfb.Bijector, ...]
        A tuple of bijectors for transforming the parameters from a constrained
        support space to an unconstrained space. Primarely used for the
        numerically stable optimization and sampling of the parameters (in a
        jit-compatible way).

    """

    def __init__(
        self,
        kernel_library: KernelLibrary,
        max_depth: int,
        num_datapoints: int,
        probs: tp.Optional[Float[jnp.ndarray, " D"]] = None,
    ) -> None:
        """
        Initialize the KernelPrior class.

        Parameters
        ----------
        kernel_library : KernelLibrary
            An instance of the KernelLibrary class, containing
            the kernel classes and operators, and transformation functions.
            The TreeStructurePrior and ParameterPrior classes are initialized
            using the kernel_library.
        max_depth : int
            The maximum depth of the nested kernel structure tree.
        num_datapoints : int
            The number of data points in the dataset.
        probs : tp.Optional[ Float[jnp.ndarray, " D"]], optional
            The probabilities of sampling each kernel (or operator) in the library.
            The array must have the same length as the the arrays in the kernel library.
            By default None, which will use a uniform distribution.
        """
        self.kernel_library = kernel_library

        self.kernel_structure_prior = TreeStructurePrior(
            self.kernel_library,
            max_depth,
            probs,
        )

        self.num_datapoints = num_datapoints

        self.parameter_prior = ParameterPrior(self.kernel_library, max_depth)

        self.graphdef = self._get_graphdef()

        self.max_kernel_parameter = (
            kernel_library.max_atom_parameters * calculate_max_leaves(self.max_depth)
        )

        self.support_bijectors = tuple(
            [
                self.kernel_library.support_transforms[tag]
                for tag in self.kernel_library.support_tag_mapping
            ]
        )

    def sample(self, key: PRNGKeyArray) -> nnx.State:
        """
        Sample a kernel structure and its parameters from the prior distribution.

        Parameters
        ----------
        key : PRNGKeyArray
            Random key for sampling.

        Returns
        -------
        nnx.State
        The state of the sampled kernel, combine with the graphdef
        to create a TreeKernel object.

        """
        tree_key, parameter_key = jr.split(key)
        kernel_structure = self.kernel_structure_prior.sample(
            tree_key,
            self.kernel_structure_prior.max_depth,
        )

        kernel = TreeKernel(
            kernel_structure,
            self.kernel_library,
            self.max_depth,
            self.num_datapoints,
        )

        _, state = nnx.split(kernel)

        kernel_state, _ = self.parameter_prior.sample(
            parameter_key,
            nnx.State(state),
        )
        return kernel_state

    def sample_kernel(self, key: PRNGKeyArray) -> TreeKernel:
        """
        Sample a kernel state using self.sample and merge
        it with the graphdef and static_state to create a
        TreeKernel.

        This function is convenient for sampling a kernel
        directly, but not vmap or jittable (at least not
        using the jax commands, potentially using the
        nnx ones.)

        Parameters
        ----------
        key : PRNGKeyArray
            Random key for sampling.

        Returns
        -------
        TreeKernel
            The sampled kernel.
        """

        kernel_state = self.sample(key)
        return nnx.merge(self.graphdef, kernel_state)

    def reconstruct_kernel(
        self,
        kernel_state: nnx.State,
    ) -> TreeKernel:
        """
        Create new TreeKernel from kernel state using the graphdef,
        and reset some kernel attributes. In principle,
        this is unnecessary since the values are already
        set in the kernel state, but we need to make it
        explicit for the jit compilation.

        Parameters
        ----------
        kernel_state : nnx.State
            The kernel state to be used to create the kernel.

        Returns
        -------
        TreeKernel
            The TreeKernel instance.
        """

        kernel = nnx.merge(self.graphdef, kernel_state)

        max_depth = self.max_depth
        max_nodes = calculate_max_nodes(max_depth)
        max_leaves = calculate_max_leaves(max_depth)
        max_stack = calculate_max_stack_size(max_depth)
        kernel.max_depth = kernel.max_depth.replace(max_depth)
        kernel.max_nodes = kernel.max_nodes.replace(max_nodes)
        kernel.max_leaves = kernel.max_leaves.replace(max_leaves)
        kernel.max_stack = kernel.max_stack.replace(max_stack)
        kernel.num_atoms = kernel.num_atoms.replace(self.kernel_library.num_atoms)
        kernel.num_datapoints = kernel.num_datapoints.replace(self.num_datapoints)
        return kernel

    def _get_graphdef(
        self,
    ) -> nnx.GraphDef:
        """
        Create a random kernel and return the graphdef.

        This function is used to create a graphdef for the
        kernel (which is set as attribute in the __init__).
        The graphdef should be the same for all possible
        kernels, so we can reuse it whenever we need to
        create a KernelTree object from a kernel state.

        Returns
        -------
        nnx.GraphDef
            The graph definition of the kernel.
        """

        key = jr.PRNGKey(42)

        kernel_structure = self.kernel_structure_prior.sample(
            key,
            self.kernel_structure_prior.max_depth,
        )

        kernel = TreeKernel(
            kernel_structure,
            self.kernel_library,
            self.max_depth,
            self.num_datapoints,
        )

        graphdef, _ = nnx.split(kernel)
        return graphdef

    @property
    def atoms(self) -> tp.List[AbstractAtom]:
        return self.kernel_library.atoms

    @property
    def operators(self) -> tp.List[AbstractOperator]:
        return self.kernel_library.operators

    @property
    def is_operator(self) -> Bool[jnp.ndarray, " D"]:
        return self.kernel_library.is_operator

    @property
    def probs(self) -> Float[jnp.ndarray, " D"]:
        return self.kernel_structure_prior.probs

    @property
    def max_depth(self) -> int:
        return self.kernel_structure_prior.max_depth

    @property
    def prior_transforms(self) -> FrozenDict[str, tfb.Bijector]:
        return self.kernel_library.prior_transforms


@partial(
    jit,
    static_argnames=(
        "max_leaves",
        "max_atom_parameters",
        "forward_bijectors",
    ),
)
def sample_parameters(
    key: PRNGKeyArray,
    state: nnx.State,
    considered_nodes: Int[jnp.ndarray, " D"],
    num_parameter_array: Int[jnp.ndarray, " D"],
    max_leaves: int,
    max_atom_parameters: int,
    support_mapping_array: Int[jnp.ndarray, "M N"],
    forward_bijectors: tuple[tp.Callable, ...],
) -> tuple[nnx.State, ScalarFloat]:
    """
    A function to sample new parameters for a state, and apply them to
    the kernel. Also returns the log probability of the sampled parameters.

    The parameters are sampled from a standard normal. Whether a
    parameter is sampled or not is determined by the tree expression
    and the leaf level map (See 'gallifrey.kernels.tree' for more
    information).
    The parameter are transformed to the desired prior distribution
    using the support_mapping_array and the forward_bijectors.

    We include an input 'considered_nodes' to only sample a specific
    subset of nodes. This is useful for structure moves, where only
    a subset of the parameters are changed. (Since arrays need to
    be of fixed shape, consider_nodes can be padded with -1 or
    any value that's not a valid leaf index.)

    Inactive parameters are set to -1.

    Parameters
    ----------
    key : PRNGKeyArray
        Random key for sampling.
    state : nnx.State
        The original kernel state to be filled with new parameters.
        The state must contain the following attributes:
        - tree_expression: The tree expression that describes the kernel structure.
        - leaf_level_map: A array that maps the index of the leaf parameter array
    considered_nodes : Int[jnp.ndarray, " D"]
        This array needs to contain all the indices that are
        supposed to be sampled. For first time sampling, that
        should be all indices (or at least all leaves). For
        structure moves (e.g. subtree-replace move), it should
        contain the indices that have been changed from the
        previous tree.
    num_parameter_array : Int[jnp.ndarray, " D"]
        An array that contains the number of parameters for each
        atom in the kernel structure. Needs to be in same order
        as the atom library.
    max_leaves : int
        The maximum number of leaves in the tree.
    max_atom_parameters : int
        The maximum number of parameters any atom will take.
    support_mapping_array : Int[jnp.ndarray, "M N"]
        An array that maps the support of the parameters to the
        corresponding bijector index.
    forward_bijectors : tuple[tp.Callable, ...]
        A tuple of bijectors to transform the sampled parameters from
        a standard normal to the desired prior distribution.

    Returns
    -------
    nnx.State
        The state with the sampled parameters.
    ScalarFloat
        The (total) log probability of the sampled parameters, sum
        of all individual log probabilities.

    """

    tree_expression: jnp.ndarray = state.tree_expression.value  # type: ignore
    leaf_level_map: jnp.ndarray = state.leaf_level_map.value  # type: ignore

    def process_params(
        state: tuple[Float[jnp.ndarray, "M N"], Float[jnp.ndarray, ""], PRNGKeyArray],
        indices: Int[jnp.ndarray, "2"],
    ) -> tuple[
        tuple[Float[jnp.ndarray, "M N"], Float[jnp.ndarray, ""], PRNGKeyArray],
        Int[jnp.ndarray, "2"],
    ]:
        """Process parameter based on if it's active and considered for sampling."""

        # unpack state and indices
        kernel_parameter, log_probability, key = state
        leaf_idx, parameter_idx = indices

        # check whether the parameter is active
        node_value = tree_expression[leaf_level_map[leaf_idx]]
        atom_num_parameters = num_parameter_array[node_value]

        is_active_node = leaf_level_map[leaf_idx] >= 0
        is_active_param = parameter_idx < atom_num_parameters
        is_active = is_active_node & is_active_param

        # check whether the parameter node is part of the considered nodes
        considered_for_sampling = jnp.isin(
            leaf_level_map[leaf_idx],
            considered_nodes,
        )

        def process_active_and_considered(indices_and_key: tuple) -> tuple:
            """If active and considered for sampling, sample and transform the
            parameter, and calculate the log probability."""
            leaf_idx, parameter_idx, key = indices_and_key

            # sample parameter
            key, subkey = jr.split(key)
            sampled_param = jr.normal(subkey)

            # calculate log probability (standard normal)
            log_prob_param = -0.5 * (sampled_param**2 + jnp.log(2 * jnp.pi))

            # transform parameter to desired prior distribution
            transformed_param = lax.switch(
                support_mapping_array[node_value, parameter_idx],
                forward_bijectors,
                sampled_param,
            )
            return transformed_param, log_prob_param, key

        def process_active_not_considered(indices_and_key: tuple) -> tuple:
            """If active but not considered, return the current parameter."""
            leaf_idx, parameter_idx, key = indices_and_key
            return kernel_parameter[leaf_idx, parameter_idx], 0.0, key

        def process_inactive(indices_and_key: tuple) -> tuple:
            """If inactive, return -1."""
            _, _, key = indices_and_key
            return jnp.array(-1.0), 0.0, key  # inactive parameters are set to -1

        # process parameter, get new parameter and key
        sampled_param, log_prob_param, key = lax.cond(
            is_active,
            lambda _: lax.cond(
                considered_for_sampling,
                process_active_and_considered,
                process_active_not_considered,
                (leaf_idx, parameter_idx, key),
            ),
            process_inactive,
            (leaf_idx, parameter_idx, key),
        )

        # update kernel parameter
        kernel_parameter = kernel_parameter.at[leaf_idx, parameter_idx].set(
            sampled_param
        )

        # update the log probability
        log_probability = log_probability + log_prob_param

        return (kernel_parameter, log_probability, key), indices

    # get initial kernel parameters
    kernel_parameters: jnp.ndarray = state.parameters.value  # type: ignore

    # get all parameter index combinations
    parameter_indices = jnp.indices((max_leaves, max_atom_parameters)).reshape(2, -1).T

    # scan over all parameters and update kernel parameters
    final_state, _ = lax.scan(
        process_params,
        (kernel_parameters, jnp.array(0.0), key),
        parameter_indices,
    )
    new_kernel_parameters = final_state[0]
    tot_log_prob = final_state[1]

    state["parameters"] = state["parameters"].replace(
        new_kernel_parameters,
    )  # type: ignore
    return state, tot_log_prob


@partial(
    jit,
    static_argnames=(
        "max_leaves",
        "max_atom_parameters",
        "inverse_bijectors",
    ),
)
def log_prob_parameters(
    state: nnx.State,
    considered_nodes: Int[jnp.ndarray, " D"],
    num_parameter_array: Int[jnp.ndarray, " D"],
    max_leaves: int,
    max_atom_parameters: int,
    support_mapping_array: Int[jnp.ndarray, "M N"],
    inverse_bijectors: tuple[tp.Callable, ...],
) -> ScalarFloat:
    """
    A function takes a state and calculates the log probability of the
    kernel parameters.

    It is assumed the parameters are sampled from a standard normal and
    then transformed to the desired prior distribution using the
    support_mapping_array and the forward_bijectors. In this function,
    we transform the parameters back to the standard normal and
    calculate the log probability.

    We include an input 'considered_nodes' to only sample a specific
    subset of nodes. This is useful for structure moves, where only
    a subset of the parameters are changed. (Since arrays need to
    be of fixed shape, consider_nodes can be padded with -1 or
    any value that's not a valid leaf index.)

    Parameters
    ----------
    state : nnx.State
        The original kernel state to be filled with new parameters.
        The state must contain the following attributes:
        - tree_expression: The tree expression that describes the kernel structure.
        - leaf_level_map: A array that maps the index of the leaf parameter array
    considered_nodes : Int[jnp.ndarray, " D"]
        This array needs to contain all the indices that are
        supposed to be sampled. For first time sampling, that
        should be all indices (or at least all leaves). For
        structure moves (e.g. subtree-replace move), it should
        contain the indices that have been changed from the
        previous tree.
    num_parameter_array : Int[jnp.ndarray, " D"]
        An array that contains the number of parameters for each
        atom in the kernel structure. Needs to be in same order
        as the atom library.
    max_leaves : int
        The maximum number of leaves in the tree.
    max_atom_parameters : int
        The maximum number of parameters any atom will take.
    support_mapping_array : Int[jnp.ndarray, "M N"]
        An array that maps the support of the parameters to the
        corresponding bijector index.
    inverse_bijectors : tuple[tp.Callable, ...]
        A tuple of bijectors to transform the parameters back to
        a standard normal.

    Returns
    -------
    ScalarFloat
        The (total) log probability of the parameters, sum
        of all individual log probabilities.

    """

    tree_expression: jnp.ndarray = state.tree_expression.value  # type: ignore
    leaf_level_map: jnp.ndarray = state.leaf_level_map.value  # type: ignore

    def process_params(
        state: tuple[Float[jnp.ndarray, "M N"], Float[jnp.ndarray, ""]],
        indices: Int[jnp.ndarray, "2"],
    ) -> tuple[
        tuple[Float[jnp.ndarray, "M N"], Float[jnp.ndarray, ""]],
        Int[jnp.ndarray, "2"],
    ]:
        """Process parameter based on if it's active and considered."""

        # unpack state and indices
        kernel_parameters, log_probability = state
        leaf_idx, parameter_idx = indices

        # check whether the parameter is active
        node_value = tree_expression[leaf_level_map[leaf_idx]]
        atom_num_parameters = num_parameter_array[node_value]

        is_active_node = leaf_level_map[leaf_idx] >= 0
        is_active_param = parameter_idx < atom_num_parameters
        is_active = is_active_node & is_active_param

        # check whether the parameter node is part of the considered nodes
        considered_for_sampling = jnp.isin(
            leaf_level_map[leaf_idx],
            considered_nodes,
        )

        def process_active_and_considered(indices: tuple) -> ScalarFloat:
            """If active and considered for sampling, transform parameter back
            to standard normal and calculate probability."""
            leaf_idx, parameter_idx = indices

            # select parameter and transform back to standard normal
            param = kernel_parameters[leaf_idx, parameter_idx]

            transformed_param = lax.switch(
                support_mapping_array[node_value, parameter_idx],
                inverse_bijectors,
                param,
            )

            # calculate log probability (standard normal)
            log_prob_param = -0.5 * (transformed_param**2 + jnp.log(2 * jnp.pi))

            return log_prob_param

        def process_others(indices: tuple) -> ScalarFloat:
            """If not active or not considered for sampling, don't
            consider for log probability."""
            return jnp.array(0.0)

        # process parameter, get probabilities
        log_prob_param = lax.cond(
            is_active & considered_for_sampling,
            process_active_and_considered,
            process_others,
            (leaf_idx, parameter_idx),
        )

        # update the log probability
        log_probability += log_prob_param

        return (kernel_parameters, log_probability), indices

    # get initial kernel parameters
    kernel_parameters: jnp.ndarray = state.parameters.value  # type: ignore

    # get all parameter index combinations
    parameter_indices = jnp.indices((max_leaves, max_atom_parameters)).reshape(2, -1).T

    # scan over all parameters and get probabilities
    parameters_and_total_log_prob, _ = lax.scan(
        process_params,
        (kernel_parameters, jnp.array(0.0)),
        parameter_indices,
    )

    return parameters_and_total_log_prob[1]


class ParameterPrior:
    """
    This class defines a prior distribution over kernel parameters.

    Attributes
    ----------
    num_parameter_array : Int[jnp.ndarray, " D"]
        An array that contains the number of parameters for each
        atom in the kernel structure. Needs to be in same order
        as the atom library.
    max_atom_parameters : int
        The maximum number of parameters for an atom in the kernel structure.
    max_leaves : int
        The maximum number of leaves in the tree.
    max_nodes : int
        The maximum number of nodes in the kernel structure.
    support_mapping_array : Int[jnp.ndarray, " D"]
        An array that maps the support of the parameters to the
        corresponding bijector index.
    forward_bijectors : tuple[tfb.Callable]
        A tuple of bijectors to transform the sampled parameters from
        a standard normal to the desired prior distribution.
    inverse_bijectors : tuple[tfb.Callable]
        A tuple of bijectors to transform the sampled parameters from
        the prior distribution back to a standard normal.

    """

    def __init__(
        self,
        kernel_library: KernelLibrary,
        max_depth: int,
    ):
        """
        Initialize the ParameterPrior class.

        Parameters
        ----------
        kernel_library : KernelLibrary
            An instance of the KernelLibrary class, containing
            the kernel classes and operators, and transformation functions.
            The prior_transforms dictionary is used to transform the
            sampled parameters to the desired prior distribution.
        max_depth : int
            The maximum depth of the kernel tree.

        """
        self.num_parameter_array = jnp.array(
            [atom.num_parameter for atom in kernel_library.library]
        )

        self.max_atom_parameters = kernel_library.max_atom_parameters
        self.max_leaves = int(calculate_max_leaves(max_depth))
        self.max_nodes = int(calculate_max_nodes(max_depth))

        self.support_mapping_array = kernel_library.support_mapping_array

        # create tuples of callables for forward and inverse bijectors
        self.forward_bijectors = tuple(
            [
                kernel_library.prior_transforms[tag].forward  # type: ignore
                for tag in kernel_library.support_tag_mapping
            ]
        )

        self.inverse_bijectors = tuple(
            [
                kernel_library.prior_transforms[tag].inverse  # type: ignore
                for tag in kernel_library.support_tag_mapping
            ]
        )

    def sample(
        self,
        key: PRNGKeyArray,
        state: nnx.State,
    ) -> tuple[nnx.State, ScalarFloat]:
        """
        Sample kernel parameter and assign it to the kernel. Also
        returns the log probability of the sampled parameters.


        The kernel parameter are sampled from a standard normal and
        transformed to follow their corresponding prior distributions
        defined by the parameter_transforms dictionary.

        Parameters
        ----------
        key : PRNGKeyArray
            Random key for sampling.
        state : nnx.State
            The original kernel state to be filled with new parameters.

        Returns
        -------
        nnx.State
            The state with the sampled parameters.
        ScalarFloat
            The log probability of the sampled parameters.

        """
        return self.sample_subset(
            key,
            state,
            jnp.arange(self.max_nodes),  # all nodes are considered
        )

    def sample_subset(
        self,
        key: PRNGKeyArray,
        state: nnx.State,
        considered_nodes: Int[jnp.ndarray, " D"],
    ) -> tuple[nnx.State, ScalarFloat]:
        """
        Sample kernel parameter and assign it to the kernel.
        Same as 'sample' method but with additional
        parameter 'considered_nodes', which can be used
        to only sample parameters for a subset of nodes.

        Parameters
        ----------
        key : PRNGKeyArray
            Random key for sampling.
        state : nnx.State
            The original kernel state to be filled with new parameters.
        considered_nodes : Int[jnp.ndarray, " D"]
            An array that contains the indices of the nodes that
            are supposed to be sampled. (Padded with -1 if
            necessary.)

        Returns
        -------
        nnx.State
            The state with the sampled parameters.
        ScalarFloat
            The log probability of the sampled parameters.
        """

        new_state, log_prob = sample_parameters(
            key,
            state,
            considered_nodes,
            self.num_parameter_array,
            self.max_leaves,
            self.max_atom_parameters,
            self.support_mapping_array,
            self.forward_bijectors,
        )
        return new_state, log_prob

    def log_prob(
        self,
        state: nnx.State,
    ) -> ScalarFloat:
        """
        Compute the log probability of the kernel parameters.

        Parameters
        ----------
        state : nnx.State
            The kernel state with the parameters.

        Returns
        -------
        ScalarFloat
            The log probability of the kernel parameters.
        """

        return self.log_prob_subset(
            state,
            jnp.arange(self.max_nodes),  # all nodes are considered
        )

    def log_prob_subset(
        self,
        state: nnx.State,
        considered_nodes: Int[jnp.ndarray, " D"],
    ) -> ScalarFloat:
        """
        Compute the log probability of the kernel parameters.
        Same as 'log_prob' method but with additional
        parameter 'considered_nodes', which can be used
        to only calculate the log probability for a subset of nodes.

        Parameters
        ----------
        state : nnx.State
            The kernel state with the parameters.
        considered_nodes : Int[jnp.ndarray, " D"]
            An array that contains the indices of the nodes that
            are supposed to be sampled. (Padded with -1 if
            necessary.)

        Returns
        -------
        ScalarFloat
            The log probability of the kernel parameters.
        """

        return log_prob_parameters(
            state,
            considered_nodes,
            self.num_parameter_array,
            self.max_leaves,
            self.max_atom_parameters,
            self.support_mapping_array,
            self.inverse_bijectors,
        )


class TreeStructurePrior(tfd.Distribution):
    """
    A prior distribution over kernel structures.

    The TreeStructurePrior is a distribution over kernel structures. The kernel
    structure is represented as a tree, where each leaf corresponds to a kernel. The
    tree is constructed by sampling from a library of kernel classes and operators.
    The operators are used to construct nested kernel structures.

    Attributes
    ----------
    library : tp.List[tp.Type[AbstractKernel] | tp.Callable]
        A list of kernel classes and operators, as defined in the KernelLibrary class.
        See gallifrey.kernels.library.KernelLibrary for more details.
    is_operator : tp.List[bool]
        A list of booleans indicating whether each element in the library is an
        operator.
    max_depth : ScalarInt
        The maximum depth of the nested kernel structure tree.
    probs : jnp.ndarray
        The probabilities of sampling each kernel (or operator) in the library.
        See kernels.library.KernelLibrary and gallifrey.config.GPConfig for more
        details.

    """

    def __init__(
        self,
        kernel_library: KernelLibrary,
        max_depth: int,
        probs: tp.Optional[Float[jnp.ndarray, " D"]] = None,
        *,
        validate_args: tp.Optional[bool] = None,
    ):
        """
        Initialize the KernelPrior distribution.

        Parameters
        ----------
        kernel_library : KernelLibrary
            An instance of the KernelLibrary class, containing the kernel classes and
            operators. (See gallifrey.kernels.library.KernelLibrary)
        max_depth : int
            The maximum depth of the nested kernel structure tree.
        probs :  Float[jnp.ndarray, " D"], optional
            The probabilities of sampling each kernel (or operator) in the library.
            The array must have the same length as the the arrays in the kernel library.
            By default None, which will use a uniform distribution.
        validate_args : tp.Optional[bool], optional
            Whether to validate input, by default None. NOT IMPLEMENTED.

        Raises
        ------
        ValueError
            If the probs are not one-dimensional.
        ValueError
            If the probs do not have the same length as the kernel_library
            along the last dimension.
        """
        # initialize the kernel functions and the probabilities
        self.library = kernel_library.library
        self.is_operator = kernel_library.is_operator
        self.max_depth = max_depth

        self.probs = probs if probs is not None else jnp.ones(len(self.library))
        if jnp.ndim(self.probs) != 1:
            raise ValueError("'probs' must be one-dimensional.")
        if not len(self.library) == len(self.probs):
            raise ValueError(
                "'probs' must have same length along the last "
                "dimension as 'kernel_library'. "
                f"Got 'probs' shape: {jnp.shape(self.probs)}, "
                f"'kernel_library' length: {len(self.library)}."
            )

    def sample_single(
        self,
        key: PRNGKeyArray,
        max_depth: tp.Optional[ScalarInt] = None,
        root_idx: ScalarInt = 0,
    ) -> Int[jnp.ndarray, " D"]:
        """
        Construct an abstract representation of a kernel structure by sampling from
        the kernel library.
        The kernels are sampled from the library according to the defined
        probabilities.
        The structure is represented as a one-dimensional array, according to
        the level-order traversal of the tree. This means:
        - The root node is at position 0.
        - The left child of a node at position i is at position 2*i + 1.
        - The right child of a node at position i is at position 2*i + 2.
        - Empty nodes are labeled -1.
        See gallifrey.kernels.tree.TreeKernel for more details and an example.

        Parameters
        ----------
        key : PRNGKeyArray
            Random key for sampling.
        max_depth : ScalarInt, optional
            The maximum depth of the nested kernel structure tree, by default None.
            If None, the maximum depth is set to the value of self.max_depth.
        root_idx : ScalarInt, optional
            The index of the root node in the tree, by default 0. Used
            for sub-trees.

        Returns
        -------
        Int[jnp.ndarray, " D"]
            An array that describes the kernel structure.
        """
        max_depth = max_depth if max_depth is not None else self.max_depth

        if max_depth < 0:
            raise ValueError("'max_depth' must be 0 or larger.")

        max_nodes = calculate_max_nodes(max_depth)

        # create sample array to be filled, this will be the output (empty
        # nodes are labeled -1)
        sample = jnp.full(max_nodes, -1)
        # create initial stack: empty except for the root node
        initial_stack = jnp.copy(sample).at[0].set(root_idx)

        pointer = 0  # initial position of the stack pointer

        initial_state = (key, sample, initial_stack, pointer)

        return _sample_single(
            initial_state,
            self.probs,
            self.is_operator,
            max_depth,
        )

    def log_prob_single(
        self,
        value: Int[jnp.ndarray, " D"],
        root_idx: ScalarInt = 0,
        path_to_hole: tp.Optional[Int[jnp.ndarray, " D"]] = None,
        hole_idx: tp.Optional[ScalarInt] = None,
    ) -> ScalarFloat:
        """
        Compute the log probability of a given kernel structure,
        as represented by the level-order tree array.
        The maximum depth of the tree is inferred from the length of the array.

        The function can also be used to calculate the log probability of a
        scaffold structure (used by the detach-attach move), by providing
        the path to the hole and the hole index.

        Parameters
        ----------
        value :  Int[jnp.ndarray, " D"]
            An array that describes the kernel structure.
        root_idx : ScalarInt, optional
            The index of the root node in the tree, by default 0.
        path_to_hole :  Int[jnp.ndarray, " D"], optional
            The path to the hole in the tree. A list of indices that
            describe the path from the root to the hole (level order
            indices), by default None (sets it to an empty array).
        hole_idx : ScalarInt, optional
            The index of the hole in the tree (level order index),
            by default None(sets it to -1 which should never be
            reached).

        Returns
        -------
        ScalarFloat
            The log probability of the given kernel structure or scaffold.
        """
        max_nodes = len(value)
        max_depth = calculate_max_depth(max_nodes)

        initial_log_p = 0.0
        inital_stack = jnp.full(max_nodes, -1).at[0].set(root_idx)
        pointer = 0
        initial_state = (initial_log_p, inital_stack, pointer)

        # set the path to the hole and the hole index, if not given
        path_to_hole = path_to_hole if path_to_hole is not None else jnp.array([])
        hole_idx = hole_idx if hole_idx is not None else -1

        return _log_prob_single(
            initial_state,
            value,
            self.probs,
            self.is_operator,
            max_depth,
            path_to_hole,
            hole_idx,
        )

    def sample(
        self,
        key: PRNGKeyArray,
        max_depth: tp.Optional[ScalarInt] = None,
        root_idx: ScalarInt = 0,
        sample_shape: tuple = (),
    ) -> Int[jnp.ndarray, "..."]:
        """
        Sample kernel structures. For details, see `sample_single`.

        Parameters
        ----------
        key : PRNGKey
            Random key for sampling.
        max_depth : ScalarInt
            The maximum depth of the nested kernel tree structure. If None,
            the maximum depth is set to the value of self.max_depth.
        root_idx : ScalarInt, optional
            The index of the root node in the tree, by default 0.
            Used for sub-trees.
        sample_shape : tuple, optional
            The sample shape for the distribution, by default ().

        Returns
        -------
        Int[jnp.ndarray, "..."]
            An array of samples describing kernel structures,
            of shape sample_shape + (max_nodes,).
        """
        max_depth = max_depth if max_depth is not None else self.max_depth

        if not sample_shape:
            return self.sample_single(key, max_depth, root_idx)

        max_nodes = calculate_max_nodes(max_depth)

        # flatten the sample_shape for vectorized sampling
        num_samples = jnp.prod(jnp.array(sample_shape))
        keys = jr.split(key, num=int(num_samples))

        samples = vmap(self.sample_single, in_axes=(0, None, None))(
            keys,
            max_depth,
            root_idx,
        )
        return samples.reshape(sample_shape + (max_nodes,))

    def log_prob(
        self,
        value: Int[jnp.ndarray, "..."],
        root_idx: ScalarInt = 0,
        path_to_hole: tp.Optional[Int[jnp.ndarray, "..."]] = None,
        hole_idx: tp.Optional[ScalarInt] = None,
    ) -> Float[jnp.ndarray, "..."]:
        """
        Compute the log probability of given kernel structures or scaffolds.
        (See log_prob_single for details.)

        Parameters
        ----------
        value :  Int[jnp.ndarray, "..."]
            Expression of kernel structure(s).
        root_idx : ScalarInt, optional
            The index of the root node in the tree, by default 0.
        path_to_hole :  Int[jnp.ndarray, "..."], optional
            Optional path to the hole in the tree, if calculating the log
            probability of a scaffold, by default None.
        hole_idx : ScalarInt, optional
            The index of the hole in the tree if calculating the log
            probability of a scaffold, by default None.

        Returns
        -------
         Float[jnp.ndarray, "..."]
            The log probability of the given kernel structures.
        """
        sample_shape = jnp.shape(value)[:-1]

        num_samples = jnp.prod(jnp.array(sample_shape))
        log_probs = vmap(self.log_prob_single, in_axes=(0, None, None, None))(
            value.reshape(int(num_samples), -1),
            root_idx,
            path_to_hole,
            hole_idx,
        )
        return jnp.asarray(log_probs).reshape(sample_shape)


@jit
def _sample_single(
    initial_state: tuple,
    probs: Float[jnp.ndarray, " D"],
    is_operator: Bool[jnp.ndarray, " D"],
    max_depth: ScalarInt,
) -> Int[jnp.ndarray, " D"]:
    """
    Construct an abstract representation of a kernel structure by sampling from
    the kernel library.
    The kernels are sampled from the library according to the defined
    probabilities.
    The structure is represented as a one-dimensional array, according to
    the level-order traversal of the tree. This means:
    - The root node is at position 0.
    - The left child of a node at position i is at position 2*i + 1.
    - The right child of a node at position i is at position 2*i + 2.
    - Empty nodes are labeled -1.
    See gallifrey.kernels.tree.TreeKernel for more details and an example.

    NOTE: This function is should be called via the sample_single method
    of the KernelPrior class.

    Parameters
    ----------
    initial_state : tuple
        Initial state for the while loop.
    probs :  Float[jnp.ndarray, " D"]
        The probabilities of sampling each kernel in the library.
    is_operator : Bool[jnp.ndarray, " D"]
        An array that indicates whether each kernel in the library is an operator.
    max_depth : ScalarInt
        The maximum depth of the nested kernel structure tree.

    Returns
    -------
     Int[jnp.ndarray, " D"]
        An array that describes the kernel structure.
    """

    def traverse_tree(
        state: tuple[PRNGKeyArray, jnp.ndarray, jnp.ndarray, ScalarInt]
    ) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, ScalarInt]:
        """Traverse the tree and sample the kernel structure."""
        key, sample, stack, pointer = state

        # pop the top of the stack
        pos = stack[pointer]
        pointer = pointer - 1

        # compute the depth of the current node
        depth = get_depth(pos)

        # determine if operators are allowed at this depth
        is_at_max_depth = depth >= max_depth
        # mask the probabilities: if at max_depth, disable operator choices
        masked_probs = jnp.where(
            is_at_max_depth,
            jnp.where(is_operator, 0.0, probs),  # if at max depth, mask operators
            probs,  # else use full probs
        )

        # create subkey for sampling and next iteration
        key, new_key = jr.split(key)

        # sample choice from masked_probs
        choice = jr.categorical(
            new_key,
            logits=jnp.log(masked_probs),
        )
        sample_updated = sample.at[pos].set(choice)

        # determine if choice is an operator
        is_op = is_operator[choice]

        def push(
            stack_and_pointer: tuple[jnp.ndarray, ScalarInt]
        ) -> tuple[jnp.ndarray, ScalarInt]:
            """If the node is an operator, push the new
            child nodes onto the stack.
            """
            stack, pointer = stack_and_pointer

            left_child = get_child_idx(pos, "l")
            right_child = get_child_idx(pos, "r")

            stack_updated = (
                stack.at[pointer + 1].set(right_child).at[pointer + 2].set(left_child)
            )
            pointer_updated = pointer + 2
            return stack_updated, pointer_updated

        def dont_push(
            stack_and_pointer: tuple[jnp.ndarray, ScalarInt]
        ) -> tuple[jnp.ndarray, ScalarInt]:
            """If the node is a leaf node, do not modify the stack."""
            return stack_and_pointer

        # Conditionally sample the node or set it as empty based on parent operator
        stack_updated, pointer_updated = lax.cond(
            is_op,
            push,
            dont_push,
            operand=(stack, pointer),
        )

        return (
            new_key,
            sample_updated,
            stack_updated,
            pointer_updated,
        )

    def condition(
        state: tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, ScalarInt]
    ) -> ScalarBool:
        """Loop condition: Continue until stack is empty."""
        _, _, _, pointer = state
        return pointer >= 0

    # execute the while loop
    _, final_sample, _, _ = lax.while_loop(condition, traverse_tree, initial_state)
    return final_sample


@jit
def _log_prob_single(
    initial_state: tuple,
    value: Int[jnp.ndarray, " D"],
    probs: Float[jnp.ndarray, " D"],
    is_operator: Bool[jnp.ndarray, " D"],
    max_depth: ScalarInt,
    path_to_hole: Int[jnp.ndarray, " D"],
    hole_idx: ScalarInt,
) -> ScalarFloat:
    """
    Compute the log probability of a given kernel structure.

    NOTE: This function is should be called via the log_prob_single method.

    Parameters
    ----------
    initial_state : tuple
        Initial state for the while loop.
    value :  Int[jnp.ndarray, " D"]
        An array that describes the kernel structure.
    probs :  Float[jnp.ndarray, " D"]
        The probabilities of sampling each kernel in the library.
    is_operator : Bool[jnp.ndarray, " D"]
        An array that indicates whether each kernel in the library is an operator.
    max_depth : ScalarInt
        The maximum depth of the nested kernel structure tree.
    path_to_hole :  Int[jnp.ndarray, " D"]
        The path to the hole in the tree.
    hole_idx : ScalarInt
        The index of the hole in the tree.

    Returns
    -------
    ScalarFloat
        The log probability of the given kernel structure.
    """

    def traverse_tree(
        state: tuple[ScalarFloat, Int[jnp.ndarray, " D"], ScalarInt]
    ) -> tuple[ScalarFloat, Int[jnp.ndarray, " D"], ScalarInt]:
        log_p, stack, pointer = state

        # pop the top of the stack
        pos = stack[pointer]
        pointer = pointer - 1

        # compute the depth of the current node
        depth = get_depth(pos)

        # get correspoding value from the kernel structure
        choice = value[pos]

        # determine if operators are allowed at this depth
        is_at_max_depth = depth >= max_depth
        # mask the probabilities: if at max_depth, disable operator choices
        masked_probs = jnp.where(
            is_at_max_depth,
            jnp.where(is_operator, 0.0, probs),  # if at max depth, mask operators
            probs,  # else use full probs
        )

        # determine if the current node is on path to the hole
        is_on_path = jnp.isin(pos, path_to_hole)
        masked_probs = jnp.where(
            is_on_path,
            jnp.where(
                ~is_operator, 0.0, masked_probs
            ),  # if on path, mask non-operators
            masked_probs,
        )
        masked_probs /= jnp.sum(masked_probs)

        # normalize masked_probs
        masked_probs /= jnp.sum(masked_probs)

        # update the log probability
        is_hole = pos == hole_idx
        log_p_updated = jnp.where(
            is_hole,
            log_p,
            log_p + jnp.log(masked_probs[choice]),
        )

        # determine if the choice is an operator node
        is_op = jnp.where(is_hole, False, is_operator[choice])

        def push(stack_and_pointer: tuple) -> tuple:
            """If the node is an operator, push the new
            child nodes onto the stack.
            """
            stack, pointer = stack_and_pointer

            left_child = get_child_idx(pos, "l")
            right_child = get_child_idx(pos, "r")

            stack_updated = (
                stack.at[pointer + 1].set(right_child).at[pointer + 2].set(left_child)
            )
            pointer_updated = pointer + 2
            return stack_updated, pointer_updated

        def dont_push(stack_and_pointer: tuple) -> tuple:
            """If the node is a leaf node, do not modify the stack."""
            return stack_and_pointer

        # update the stack based on whether the current node is an operator
        new_stack, new_pointer = lax.cond(
            is_op,
            push,
            dont_push,
            operand=(stack, pointer),
        )

        return (log_p_updated, new_stack, new_pointer)

    def condition(
        state: tuple[ScalarFloat, Int[jnp.ndarray, " D"], ScalarInt]
    ) -> ScalarBool:
        """Loop condition: Continue until stack is empty."""
        _, _, pointer = state
        return pointer >= 0

    # iterate over the sample using lax.while_loop
    final_log_p, _, _ = lax.while_loop(
        condition,
        traverse_tree,
        initial_state,
    )
    return jnp.array(final_log_p)



from functools import partial

import jax.numpy as jnp
import jax.random as jr
from flax import nnx
from jax import debug, jit, lax
from jaxtyping import Bool, Float, Int, PRNGKeyArray

from gallifrey.kernels.prior import KernelPrior
from gallifrey.moves.subtree_replace import replace_subtree_structure_with
from gallifrey.parameter import move_parameters
from gallifrey.utils.tree_helper import (
    calculate_max_nodes,
    generate_random_path,
    get_child_idx,
    get_depth,
    get_subtree,
    pick_random_node,
    reconstruct_path,
)
from gallifrey.utils.typing import ScalarBool, ScalarFloat, ScalarInt


@partial(jit, static_argnames=("kernel_prior", "verbosity"))
def detach_attach_move(
    key: PRNGKeyArray,
    kernel_state: nnx.State,
    kernel_prior: KernelPrior,
    detach_prob: ScalarFloat = 0.5,
    verbosity: int = 0,
) -> tuple[nnx.State, ScalarFloat]:
    """
    Perform detach-attach move. In this move, a subtree is detached
    from the tree expression (at idx a), and a scaffold is attached
    to the tree expression (at idx a). The subtree is then reattached
    to the scaffold.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    kernel_state : nnx.State
        The original kernel state, must have attributes:
        - parameters
        - leaf_level_map
        - node_sizes
        - tree_expression
        - is_operator
    kernel_prior : KernelPrior
        The kernel prior object that contains the kernel structure prior.
    detach_prob : ScalarFloat, optional
        The probability of performing the detach move, by default 0.5.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information is printed
        if verbosity > 2.

    Returns
    -------
    nnx.State
        The updated kernel state after the move.
    ScalarFloat
        The log acceptance ratio contribution of the move,
        needs to be added to the log acceptance ratio of the
        subtree-replace move.

    """
    key, selection_subkey, move_subkey = jr.split(key, 3)

    # Cannot perform detach-attach move on kernels with max depth = 1.
    perform_detach = jr.bernoulli(
        selection_subkey,
        jnp.where(
            kernel_state.node_sizes.value[0] == 1,
            jnp.array(0.0),
            jnp.asarray(detach_prob),
        ),
    )
    if verbosity > 2:
        lax.cond(
            perform_detach,
            lambda _: debug.print("Performing detach move."),
            lambda _: debug.print("Performing attach move."),
            perform_detach,
        )
    # ratio of the probabilities of proposing detach and attach moves,
    # needed for acceptance ratio calculation (Saad2023 - Proposition 2)
    log_detach_vs_attach_prob = jnp.log(1 - detach_prob) - jnp.log(detach_prob)

    # perform detach or attach move
    new_kernel_state, log_q_D, log_q_A = lax.cond(
        perform_detach,
        lambda key: detach_move(key, kernel_state, kernel_prior, verbosity=verbosity),
        lambda key: attach_move(key, kernel_state, kernel_prior, verbosity=verbosity),
        move_subkey,
    )

    # calculate the log acceptance ratio contribution for the move
    # (Saad2023 - Proposition 2)
    log_acceptance_ratio_contribution = lax.cond(
        perform_detach,
        lambda log_p: log_p,
        lambda log_p: -log_p,
        log_q_A - log_q_D + log_detach_vs_attach_prob,
    )

    return new_kernel_state, log_acceptance_ratio_contribution


@partial(jit, static_argnames=("kernel_prior", "verbosity"))
def attach_move(
    key: PRNGKeyArray,
    kernel_state: nnx.State,
    kernel_prior: KernelPrior,
    verbosity: int = 0,
) -> tuple[nnx.State, ScalarFloat, ScalarFloat]:
    """
    Perform attach move. In this move, a subtree is detached from the
    tree expression (at idx a), a scaffold is attached to the tree
    expression (at idx a), and the subtree originally at idx a is
    reattached to the scaffold (at idx b).

    Also returns the probabilities associated with the proposals q_A
    (going from the original tree to the new tree using the attach move)
    and q_D (going from the new tree to the original tree using a detach move).

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    kernel_state : nnx.State
        The original kernel state, must have attributes:
        - parameters
        - leaf_level_map
        - node_sizes
        - tree_expression
        - is_operator
    kernel_prior : KernelPrior
        The prior to sample the new subtree from.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information is printed
        if verbosity > 2.


    Returns
    -------
    nnx.State
        The updated kernel state after the move.
    ScalarFloat
        The probability associated with the proposal q_D (b|a, k', theta'),
        i.e. the transition probability from the new tree to the original tree
        using a detach move.
    ScalarFloat
        The probability associated with the proposal q_A (b|a, k, theta),
        i.e. going from the original tree to the new tree using the attach move.

    """
    key, parameter_key, structure_key = jr.split(key, 3)

    # get the new structure and the mapping between the old and new indices
    new_structure, changes, index_mapping, log_p_path, log_p_scaffold, idx_a, _ = (
        structure_attach_move(
            structure_key,
            kernel_state.tree_expression.value,  # type: ignore
            kernel_state.post_level_map.value,  # type: ignore
            kernel_state.node_heights.value,  # type: ignore
            kernel_prior,
            verbosity=verbosity,
        )
    )

    # get attributes of new tree
    new_kernel = nnx.merge(kernel_prior.graphdef, kernel_state)
    new_kernel.init_tree(
        new_structure,
        kernel_prior.kernel_library,
        kernel_prior.max_depth,
        kernel_prior.num_datapoints,
    )

    # move parameters of subtree from old indices to new indices
    _, new_kernel_state = nnx.split(new_kernel)
    new_kernel_state = move_parameters(
        new_kernel_state,
        index_mapping.T,
        kernel_prior.max_depth,
    )

    # sample new kernel parameters, but first remove the nodes
    # from changes that were already moved to a new place (since we
    # want to keep the parameters of the moved nodes and only sample
    # parameters for completely new nodes introduced by scaffold)
    changes = jnp.where(
        jnp.isin(changes, index_mapping[1]),
        -1,
        changes,
    )

    new_kernel_state, log_prob_parameter = kernel_prior.parameter_prior.sample_subset(
        parameter_key,
        new_kernel_state,
        considered_nodes=changes,
    )
    # calculate q_A (b|a, k, theta), i.e. probability associated with the
    # proposal going from the original tree to the new tree using the attach move
    log_q_A = log_p_path + log_p_scaffold + log_prob_parameter

    # calculate q_D (b|a, k', theta'), i.e. the transition probability
    # from the new tree to the original tree using a detach move
    log_q_D = jnp.log(1 / new_kernel.node_sizes.value[idx_a])

    return new_kernel_state, log_q_D, log_q_A


def structure_attach_move(
    key: PRNGKeyArray,
    tree_expression: Int[jnp.ndarray, " D"],
    post_level_map: Int[jnp.ndarray, " D"],
    node_heights: Int[jnp.ndarray, " D"],
    kernel_prior: KernelPrior,
    verbosity: int = 0,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, "2 D"],
    ScalarFloat,
    ScalarFloat,
    ScalarInt,
    ScalarInt,
]:
    """
    Perform attach move on the kernel structure array. In this move,
    a subtree is detached from the tree expression (at idx a), a
    scaffold is attached to the tree expression (at idx a), and the
    subtree originally at idx a is reattached to the scaffold (at idx b).

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    tree_expression : Int[jnp.ndarray, " D"]
        The tree expression, given in level order notation.
    post_level_map : Int[jnp.ndarray, " D"]
        The array that maps the post order index to the level order index.
        (This is useful, since the post_level_map contains all the indices
        of the nodes in the tree, which we can sample).
    node_heights : Int[jnp.ndarray, " D"]
        The heights of each node in the tree. (See
        gallifrey.kernels.tree.TreeKernel for more information.)
    kernel_prior : KernelPrior
        The kernel prior object that contains the kernel structure prior.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information is printed
        if verbosity > 2.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The new tree expression with the scaffold attached and the subtree
        reattached.
    Int[jnp.ndarray, " D"]
        The changes in the tree structure, a list of indices that differ
        between the original tree and the new tree.
    Int[jnp.ndarray, "2 D"]
        The mapping between the old and new indices of the nodes in the
        subtree. The first column contains the old indices, and the second
        column contains the new indices.
    ScalarFloat
        The log probability associated with the path to the hole in the
        scaffold.
    ScalarFloat
        The log probability associated with the scaffold.
    ScalarInt
        The index of the root of the subtree that was detached/scaffold
        attached (idx a).
    ScalarInt
        The index where the subtree was reattached (idx b).
    """
    key, idx_a_subkey, path_subkey, scaffold_subkey = jr.split(key, 4)

    # select where the scaffold will be attached
    idx_a = pick_random_node(idx_a_subkey, post_level_map)

    # generate a random path to the hole in the scaffold (this is where
    # the subtree originally at idx_a will be re-attached),
    # the path needs to respect the node height of the detached subtree,
    # since it shouldn't overshoot the max depth when reattached
    idx_b, path, path_log_prob = generate_random_path(
        path_subkey,
        idx_a,
        max_depth=kernel_prior.max_depth,
        node_height=node_heights[idx_a],
    )

    scaffold, scaffold_log_p = scaffold_proposal(
        key,
        max_depth=kernel_prior.kernel_structure_prior.max_depth,
        path_to_hole=path,
        hole_idx=idx_b,
        probs=kernel_prior.kernel_structure_prior.probs,
        is_operator=kernel_prior.kernel_structure_prior.is_operator,
    )

    # transplant the subtree from the original tree at idx_a to the scaffold
    # at idx_b
    scaffold_with_subtree, index_mapping = transplant_subtree(
        tree_expression,
        scaffold,
        idx_a,
        idx_b,
        int(calculate_max_nodes(kernel_prior.max_depth)),
    )

    # replace the subtree at idx_a with the scaffold + subtree at idx_b
    new_tree_expression, changes = replace_subtree_structure_with(
        tree_expression,
        scaffold_with_subtree,
        idx_a,
    )

    if verbosity > 2:
        debug.print("idx_a: {}", idx_a)
        debug.print("idx_b: {}", idx_b)
        debug.print("path: {}", path)
        debug.print("scaffold: {}", scaffold)
        debug.print("scaffold_with_subtree: {}", scaffold_with_subtree)
        debug.print("path_log_prob: {}", path_log_prob)
        debug.print("scaffold_log_p: {}", scaffold_log_p)

    return (
        new_tree_expression,
        changes,
        index_mapping,
        path_log_prob,
        scaffold_log_p,
        idx_a,
        idx_b,
    )


@partial(jit, static_argnames=("kernel_prior", "verbosity"))
def detach_move(
    key: PRNGKeyArray,
    kernel_state: nnx.State,
    kernel_prior: KernelPrior,
    verbosity: int = 0,
) -> tuple[nnx.State, ScalarFloat, ScalarFloat]:
    """
    Perform detach move. In this move, a scaffold is detached
    from the tree expression (at idx a), and a subtree from the
    scaffold (at idx b) is detached and then reattached to the
    root (idx a) of the scaffold.

    Also returns the probabilities associated with the proposals q_D
    (going from the original tree to the new tree using the detach move)
    and q_A (going from the new tree to the original tree using an attach
    move).

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    kernel_state : nnx.State
        The original kernel state, must have attributes:
        - parameters
        - leaf_level_map
        - node_sizes
        - tree_expression
        - is_operator
    kernel_prior : KernelPrior
        The prior to sample the new subtree from.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information is printed
        if verbosity > 2.

    Returns
    -------
    nnx.State
        The updated kernel state after the move.
    ScalarFloat
        The probability associated with the proposal q_D (b|a, k, theta),
        i.e. going from the original tree to the new tree using the detach move.
    ScalarFloat
        The probability associated with the proposal q_A (b|a, k', theta'),
        i.e. the transition probability from the new tree to the original tree
        using an attach move.

    """
    # get the new structure and the mapping between the old and new indices
    new_structure, index_mapping, idx_a, idx_b = structure_detach_move(
        key,
        kernel_state.tree_expression.value,  # type: ignore
        kernel_state.post_level_map.value,  # type: ignore
        verbosity=verbosity,
    )

    # get attributes of new tree
    new_kernel = kernel_prior.reconstruct_kernel(kernel_state)
    new_kernel.init_tree(
        new_structure,
        kernel_prior.kernel_library,
        kernel_prior.max_depth,
        kernel_prior.num_datapoints,
    )

    # move parameters from old indices to new indices
    _, new_kernel_state = nnx.split(new_kernel)
    new_kernel_state = move_parameters(
        new_kernel_state,
        index_mapping.T,
        kernel_prior.max_depth,
    )

    # calculate q_D (b|a, k, theta), i.e. the probability associated with the
    # detach move going from the original tree to the new tree using the detach move
    # (which is equal to the probability of sampling idx_b given idx_a
    # was sampled, which is simply the inverse of the number of nodes in the
    # subtree, assuming uniform sampling)
    log_q_D = jnp.log(1 / kernel_state.node_sizes.value[idx_a])  # type: ignore

    # calculate q_A (b|a, k', theta'), i.e. involution attach move
    # going from the new tree to the original tree using an attach move
    path, log_p_path = reconstruct_path(
        idx_a,
        idx_b,
        max_depth=kernel_prior.kernel_structure_prior.max_depth,
    )
    # we use the original kernel here, since we are calculating the probability
    # of sampling exactly the (now) missing scaffold, so we have to traverse
    # the scaffold from the original kernel
    log_p_scaffold = kernel_prior.kernel_structure_prior.log_prob_single(
        kernel_state.tree_expression.value,  # type: ignore
        path_to_hole=path,
        hole_idx=idx_b,
    )
    log_q_A = log_p_path + log_p_scaffold

    return new_kernel_state, log_q_D, log_q_A


def structure_detach_move(
    key: PRNGKeyArray,
    tree_expression: Int[jnp.ndarray, " D"],
    post_level_map: Int[jnp.ndarray, " D"],
    empty_node_value: ScalarInt = -1,
    verbosity: int = 0,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, "2 D"],
    ScalarInt,
    ScalarInt,
]:
    """
    Perform detach move on the kernel structure array. In this move,
    a scaffold is detached from the tree expression (at idx a), and a
    subtree from the scaffold (at idx b) is detached and then
    reattached to the root (idx a) of the scaffold.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key for sampling.
    tree_expression : Int[jnp.ndarray, " D"]
        The tree expression, given in level order notation.
    post_level_map : Int[jnp.ndarray, " D"]
        The array that maps the post order index to the level order index.
        (This is useful, since the post_level_map contains all the indices
        of the nodes in the tree, which we can sample).
    empty_node_value : ScalarInt, optional
        The value to fill the hole with in the subtree, by default -1.
        (Must be negative.)
    verbosity : int, optional
        The verbosity level, by default 0. Debug information is printed
        if verbosity > 2.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The new tree expression with the scaffold removed and the subtree
        reattached.
    Int[jnp.ndarray, "2 D"]
        The mapping between the old and new indices of the nodes in the
        subtree. The first column contains the old indices, and the second
        column contains the new indices.
    ScalarInt
        The index of the root of the scaffold (idx a).
    ScalarInt
        The index of the root of the subtree that was detached from the
        scaffold (idx b).
    """

    key, idx_a_subkey, idx_b_subkey = jr.split(key, 3)

    # choose the root of the scaffold to detach
    idx_a = pick_random_node(idx_a_subkey, post_level_map)

    # detach the subtree
    amputee_tree, subtree_at_a, subtree_at_a_indices = detach_subtree(
        tree_expression,
        idx_a,
        fill_value=empty_node_value,
    )

    # select the root of the subtree to detach from the scaffold
    idx_b = pick_random_node(idx_b_subkey, subtree_at_a_indices)

    # get the subtree from the scaffold and move it to the root of the scaffold,
    # also get mapping between old and new indices
    subtree_at_b, index_mapping = get_subtree(subtree_at_a, idx_b, new_root_idx=idx_a)

    if verbosity > 2:
        debug.print("idx_a: {}", idx_a)
        debug.print("idx_b: {}", idx_b)
        debug.print("Remaining Original Tree: {}", amputee_tree)
        debug.print("Subtree at a: {}", subtree_at_a)
        debug.print("Subtree at b: {}", subtree_at_b)

    # attach the subtree at the root of the scaffold (this assumes the empty nodes in
    # the original tree are -1, and all filled nodes are >= 0)
    new_tree = jnp.where(subtree_at_b > amputee_tree, subtree_at_b, amputee_tree)

    return new_tree, index_mapping, idx_a, idx_b


def scaffold_proposal(
    key: PRNGKeyArray,
    max_depth: ScalarInt,
    path_to_hole: Int[jnp.ndarray, " D"],
    hole_idx: ScalarInt,
    probs: Float[jnp.ndarray, " D"],
    is_operator: Bool[jnp.ndarray, " D"],
    empty_value: ScalarInt = -1,
) -> tuple[
    Int[jnp.ndarray, " D"],
    ScalarFloat,
]:
    """
    Propose the tree structure for the scaffold in the
    attach move. Also returns the log-probability of the proposal.

    The scaffold is a tree structure proposal very similar
    to the samples created from the kernel prior (see
    gallifrey.kernels.prior.TreeStructurePrior). The difference
    is that this prior is conditioned by the path to the hole, meaning
    some branches are fixed to be operator nodes, so that the hole can
    be reached.
    The root of the scaffold is assumed to be first index in the path_to_hole.

    The hole itself is fixed to be a leaf node, and empty (filled by the
    empty_value) in the scaffold.
    NOTE: This means the returned scaffold is not a valid kernel structure
    in itself, the hole must be filled by a valid subtree.

    Parameters
    ----------
    key : PRNGKeyArray
        Random key for sampling.
    max_depth : ScalarInt
        The maximum depth of the nested kernel structure tree.
    path_to_hole :  Int[jnp.ndarray, " D"]
        The path to the hole in the tree. A list of indices that
        describe the path from the root to the hole (level order
        indices).
    hole_idx : ScalarInt
        The index of the hole in the tree (level order index).
    probs :  Float[jnp.ndarray, " D"]
        The probabilities of sampling each kernel in the library.
    is_operator : Bool[jnp.ndarray, " D"]
        An array that indicates whether each kernel in the library is an operator.
    empty_value : ScalarInt, optional
        The value to fill the hole with in the scaffold, by default -1.

    Returns
    -------
     Int[jnp.ndarray, " D"]
        An array that describes the scaffold structure.
    ScalarFloat
        The log probability associated with the scaffold.
    """
    max_nodes = calculate_max_nodes(max_depth)

    # create sample array to be filled, this will be the output (empty
    # nodes are labeled -1)
    sample = jnp.full(max_nodes, -1)
    # create initial stack: empty except for the root node, start at beginning of path
    initial_stack = jnp.copy(sample).at[0].set(path_to_hole[0])
    pointer = 0  # initial position of the stack pointer
    initial_log_p = 0.0  # initial probability

    initial_state = (key, sample, initial_stack, pointer, initial_log_p)

    return _scaffold_proposal(
        initial_state,
        probs,
        is_operator,
        path_to_hole,
        hole_idx,
        max_depth,
        empty_value,
    )


@jit
def _scaffold_proposal(
    initial_state: tuple,
    probs: Float[jnp.ndarray, " D"],
    is_operator: Bool[jnp.ndarray, " D"],
    path_to_hole: Int[jnp.ndarray, " D"],
    hole_idx: ScalarInt,
    max_depth: ScalarInt,
    empty_value: ScalarInt = -1,
) -> tuple[jnp.ndarray, ScalarFloat]:
    """
    Construct a scaffold proposal for the attach move.

    NOTE: This function is should be called via the scaffold_proposal
    function.

    Parameters
    ----------
    initial_state : tuple
        Initial state for the while loop.
    probs :  Float[jnp.ndarray, " D"]
        The probabilities of sampling each kernel in the library.
    is_operator : Bool[jnp.ndarray, " D"]
        An array that indicates whether each kernel in the library is an operator.
    path_to_hole :  Int[jnp.ndarray, " D"]
        The path to the hole in the tree. A list of indices that
        describe the path from the root to the hole (level order
        indices).
    hole_idx : ScalarInt
        The index of the hole in the tree (level order index).
    max_depth : ScalarInt
        The maximum depth of the nested kernel structure tree.
    empty_value : ScalarInt, optional
        The value to fill the hole with in the scaffold, by
        default -1.


    Returns
    -------
     Int[jnp.ndarray, " D"]
        An array that describes the scaffold structure.
    ScalarFloat
        The log probability associated with the scaffold.
    """

    def traverse_tree(state: tuple) -> tuple:
        """Traverse the tree and sample the scaffold."""
        key, sample, stack, pointer, log_p = state

        # pop the top of the stack
        pos = stack[pointer]
        pointer = pointer - 1

        # compute the depth of the current node
        depth = get_depth(pos)

        # determine if operators are allowed at this depth
        is_at_max_depth = depth >= max_depth
        # mask the probabilities: if at max_depth, disable operator choices
        masked_probs = jnp.where(
            is_at_max_depth,
            jnp.where(is_operator, 0.0, probs),  # if at max depth, mask operators
            probs,  # else use full probs
        )

        # determine if the current node is on path to the hole
        is_on_path = jnp.isin(pos, path_to_hole)
        masked_probs = jnp.where(
            is_on_path,
            jnp.where(
                ~is_operator, 0.0, masked_probs
            ),  # if on path, mask non-operators
            masked_probs,
        )
        masked_probs /= jnp.sum(masked_probs)

        # create subkey for sampling and next iteration
        key, new_key = jr.split(key)

        # sample choice from masked_probs
        choice = jr.categorical(
            new_key,
            logits=jnp.log(masked_probs),
        )

        # check if the current node is the hole
        is_hole = pos == hole_idx

        # if node is the hole, set the choice to the empty value
        choice = jnp.where(
            is_hole,
            empty_value,
            choice,
        )
        sample_updated = sample.at[pos].set(choice)

        # update the probability for the scaffold
        log_p_updated = jnp.where(
            is_hole,
            log_p,
            log_p + jnp.log(masked_probs[choice]),
        )

        # determine if choice is an operator
        is_op = jnp.where(is_hole, False, is_operator[choice])

        def push(stack_and_pointer: tuple) -> tuple:
            """If the node is an operator, push the new
            child nodes onto the stack.
            """
            stack, pointer = stack_and_pointer

            left_child = get_child_idx(pos, "l")
            right_child = get_child_idx(pos, "r")

            stack_updated = (
                stack.at[pointer + 1].set(right_child).at[pointer + 2].set(left_child)
            )
            pointer_updated = pointer + 2
            return stack_updated, pointer_updated

        def dont_push(stack_and_pointer: tuple) -> tuple:
            """If the node is a leaf node, do not modify the stack."""
            return stack_and_pointer

        # Conditionally sample the node or set it as empty based on parent operator
        stack_updated, pointer_updated = lax.cond(
            is_op,
            push,
            dont_push,
            operand=(stack, pointer),
        )

        return (
            new_key,
            sample_updated,
            stack_updated,
            pointer_updated,
            log_p_updated,
        )

    def condition(state: tuple) -> ScalarBool:
        """Loop condition: Continue until stack is empty."""
        _, _, _, pointer, _ = state
        return pointer >= 0

    # execute the while loop
    _, final_sample, _, _, final_log_p = lax.while_loop(
        condition,
        traverse_tree,
        initial_state,
    )
    return final_sample, final_log_p


def detach_subtree(
    tree_expression: Int[jnp.ndarray, " D"],
    subtree_root_idx: ScalarInt,
    fill_value: ScalarInt = -1,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
]:
    """
    Cut a subtree from a tree expression, starting at the subtree root
    node.

    Returns the tree_expression with the subtree removed, the subtree
    expression, and the indices of the nodes in the subtree.

    Parameters
    ----------
    tree_expression : Int[jnp.ndarray, " D"]
        The tree expression, given in level order notation.
    subtree_root_idx : ScalarInt
        The index of node where the subtree is rooted.
    fill_value : ScalarInt, optional
        The value to fill the hole with in the subtree, by default -1.
        (Must be negative.)
    Returns
    -------
    Int[jnp.ndarray, " D"]
        The (level order) tree expression with the subtree removed.
    Int[jnp.ndarray, " D"]
        The (level order) tree expression of the subtree, same length as the
        input tree expression.
    Int[jnp.ndarray, " D"]
        The indices of the nodes in the subtree.

    """
    max_nodes = tree_expression.size

    # if tree_expression[subtree_root_idx] == -1:
    #     raise ValueError("The start index must be a non-empty node.")
    # if fill_value >= 0:
    #     raise ValueError("The fill value must be negative.")

    # create initial state
    subtree_expression = jnp.full(max_nodes, -1)
    index_array = jnp.copy(subtree_expression)
    index_pointer = 0
    stack = jnp.copy(subtree_expression).at[0].set(subtree_root_idx)
    stack_pointer = 0

    initial_state = (
        tree_expression,
        subtree_expression,
        index_array,
        index_pointer,
        stack,
        stack_pointer,
    )

    tree_expression, subtree_expression, index_array, _ = _detach_subtree(
        initial_state,
        max_nodes,
        fill_value,
    )

    return tree_expression, subtree_expression, index_array


@jit
def _detach_subtree(
    initial_state: tuple[
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        ScalarInt,
        Int[jnp.ndarray, " D"],
        ScalarInt,
    ],
    max_nodes: ScalarInt,
    fill_value: ScalarInt = -1,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
    ScalarInt,
]:
    """
    Traverse the tree expression in level-order and cut a subtree from
    the tree expression.

    NOTE: This function is not meant to be called directly. Use get_subtree
    instead.

    Parameters
    ----------
    initial_state : tuple
        The initial state of the traversal. Contains the tree expression,
        the subtree expression, the index array, the index pointer, the stack
        and the stack pointer.
    max_nodes : ScalarInt
        The maximum number of nodes in the tree.
    fill_value : ScalarInt, optional
        The value to fill the hole with in the subtree, by default -1.
        (Must be negative.)

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The (level order) tree expression with the subtree removed.
    Int[jnp.ndarray, " D"]
        The (level order) tree expression of the subtree, same length as the
        input tree expression.
    Int[jnp.ndarray, " D"]
        The indices of the nodes in the subtree.
    ScalarInt
        The number of nodes in the subtree.

    """

    def traverse(state: tuple) -> tuple:
        """Traverse the tree expression in level-order."""
        (
            tree_expression,
            subtree_expression,
            index_array,
            index_pointer,
            old_stack,
            stack_pointer,
        ) = state

        # get old and new indices
        idx = old_stack[stack_pointer]

        # fill index array with the (new) index of the current node
        index_array_updated = index_array.at[index_pointer].set(idx)
        index_pointer_updated = index_pointer + 1

        # fill the new tree expression with the current node (at the new index)
        subtree_expression_updated = subtree_expression.at[idx].set(
            tree_expression[idx]
        )
        # remove node from original tree
        tree_expression_updated = tree_expression.at[idx].set(fill_value)

        # check if the current node is a leaf
        right_child_idx = get_child_idx(idx, "r")
        is_leaf = jnp.logical_or(
            right_child_idx > max_nodes - 1,
            tree_expression[right_child_idx - 1] == -1,
        )

        def process_leaf(stacks_and_pointer: tuple) -> tuple:
            """If the current node is a leaf, move to the parent node."""
            old_stack, pointer = stacks_and_pointer
            return old_stack, pointer - 1

        def process_non_leaf(stacks_and_pointer: tuple) -> tuple:
            """If the current node is not a leaf, update the stacks
            (add the children of the current node)."""
            old_stack, pointer = stacks_and_pointer

            left_child_idx = get_child_idx(idx, "l")
            right_child_idx = get_child_idx(idx, "r")

            old_stack_updated = old_stack.at[pointer].set(left_child_idx)
            old_stack_updated = old_stack_updated.at[pointer + 1].set(right_child_idx)

            return old_stack_updated, pointer + 1

        # update the stack and pointer
        old_stack_updated, pointer_updated = lax.cond(
            is_leaf,
            process_leaf,
            process_non_leaf,
            operand=(old_stack, stack_pointer),
        )

        return (
            tree_expression_updated,
            subtree_expression_updated,
            index_array_updated,
            index_pointer_updated,
            old_stack_updated,
            pointer_updated,
        )

    def condition(state: tuple) -> ScalarBool:
        """Check if the traversal is complete."""
        _, _, _, _, _, stack_pointer = state
        return stack_pointer >= 0

    # iterate over the sample using lax.while_loop
    (
        final_tree_expression,
        final_subtree_expression,
        final_index_array,
        final_index_pointer,
        _,
        _,
    ) = lax.while_loop(
        condition,
        traverse,
        initial_state,
    )
    return (
        final_tree_expression,
        final_subtree_expression,
        final_index_array,
        final_index_pointer,
    )


@partial(jit, static_argnames=("max_nodes",))
def transplant_subtree(
    donor_tree_expression: Int[jnp.ndarray, " D"],
    recipient_tree_expression: Int[jnp.ndarray, " D"],
    donor_root_idx: ScalarInt,
    recipient_root_idx: ScalarInt,
    max_nodes: int,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, "2 D"],
]:
    """
    Transplant a subtree from one tree to another. The subtree is
    rooted at the donor root index and re-rooted at the recipient
    root index.
    The recipient tree must have an empty node at the recipient root index.

    Parameters
    ----------
    donor_tree_expression : Int[jnp.ndarray, " D"]
        The tree expression of the donor tree.
    recipient_tree_expression : Int[jnp.ndarray, " D"]
        The tree expression of the recipient tree.
    donor_root_idx : ScalarInt
        The index of the root of the subtree to be moved.
    recipient_root_idx : ScalarInt
        The index of the root of the subtree to be moved to.
    max_nodes : int
        The maximum number of nodes in the tree.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The tree expression of the recipient tree with the attached subtree.
    Int[jnp.ndarray, "2 D"]
        The indices of the moved nodes. Two rows, the first row
        contains the indices of the subtree in the donor tree, the second
        row contains the indices of the subtree in the recipient tree.

    """
    # if donor_tree_expression[donor_root_idx] == -1:
    #     raise ValueError("The donor root index must be a non-empty node.")
    # if recipient_tree_expression[recipient_root_idx] != -1:
    #     raise ValueError("The recipient root index must be an empty node.")

    # create initial state
    index_array = jnp.full((2, max_nodes), -1)
    index_pointer = 0
    stack = jnp.full(max_nodes, -1).at[0].set(donor_root_idx)
    new_stack = jnp.copy(stack).at[0].set(recipient_root_idx)
    stack_pointer = 0

    initial_state = (
        recipient_tree_expression,
        index_array,
        index_pointer,
        stack,
        new_stack,
        stack_pointer,
    )

    tree_expression, index_array, _ = _transplant_subtree(
        donor_tree_expression,
        initial_state,
        max_nodes,
    )

    index_array = index_array
    # if jnp.any(index_array >= max_nodes):
    #     raise ValueError(
    #         "The index array is out of bounds. This could happen if the tree "
    #         "is re-rooted to a spot where one of its leaves exceeds the "
    #         "maximum number of nodes."
    #     )
    return tree_expression, index_array


@jit
def _transplant_subtree(
    donor_tree_expression: Int[jnp.ndarray, " D"],
    initial_state: tuple[
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, "2 D"],
        ScalarInt,
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        ScalarInt,
    ],
    max_nodes: ScalarInt,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, "2 D"],
    ScalarInt,
]:
    """
    Traverse the tree expression in level-order, transplanting the subtree
    rooted at the donor root index to the recipient root index.

    NOTE: This function is not meant to be called directly. Use transplant_subtree
    instead.

    Parameters
    ----------
    donor_tree_expression : Int[jnp.ndarray, " D"]
        The tree expression of the donor tree.
    initial_state : tuple
        The initial state of the traversal. Contains the tree expression of the
        recipient tree, the index array, the index pointer, the stack of the
        nodes to be visited in the donor tree, the stack of the nodes to be visited
        in the recipient tree, and the stack pointer.
    max_nodes : ScalarInt
        The maximum number of nodes in the tree.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The (level order) tree expression of the recipient tree with
        the attached subtree.
    Int[jnp.ndarray, " D"]
        The indices of the moved nodes. Two rows, the first row
        contains the indices of the subtree in the donor tree, the second
        row contains the indices of the subtree in the recipient tree.
    ScalarInt
        The number of indices in the index array.

    """

    def traverse(state: tuple) -> tuple:
        """Traverse the tree expression in level-order."""
        (
            recipient_tree_expression,
            index_array,
            index_pointer,
            old_stack,
            new_stack,
            stack_pointer,
        ) = state

        # get old and new indices
        old_idx = old_stack[stack_pointer]
        new_idx = new_stack[stack_pointer]

        # fill index array
        index_array_updated = index_array.at[0, index_pointer].set(old_idx)
        index_array_updated = index_array_updated.at[1, index_pointer].set(new_idx)
        index_pointer_updated = index_pointer + 1

        # move the current node to the new position
        recipient_tree_expression_updated = recipient_tree_expression.at[new_idx].set(
            donor_tree_expression[old_idx]
        )

        # check if the current node is a leaf
        right_child_idx = get_child_idx(old_idx, "r")
        is_leaf = jnp.logical_or(
            right_child_idx > max_nodes - 1,
            donor_tree_expression[right_child_idx - 1] == -1,
        )

        def process_leaf(stacks_and_pointer: tuple) -> tuple:
            """If the current node is a leaf, move to the parent node."""
            old_stack, new_stack, pointer = stacks_and_pointer
            return old_stack, new_stack, pointer - 1

        def process_non_leaf(stacks_and_pointer: tuple) -> tuple:
            """If the current node is not a leaf, update the stacks
            (add the children of the current node)."""
            old_stack, new_stack, pointer = stacks_and_pointer

            left_child_idx = get_child_idx(old_idx, "l")
            right_child_idx = get_child_idx(old_idx, "r")

            new_left_child_idx = get_child_idx(new_idx, "l")
            new_right_child_idx = get_child_idx(new_idx, "r")

            old_stack_updated = old_stack.at[pointer].set(left_child_idx)
            old_stack_updated = old_stack_updated.at[pointer + 1].set(right_child_idx)

            new_stack_updated = new_stack.at[pointer].set(new_left_child_idx)
            new_stack_updated = new_stack_updated.at[pointer + 1].set(
                new_right_child_idx
            )

            return old_stack_updated, new_stack_updated, pointer + 1

        # update the stack and pointer
        old_stack_updated, new_stack_updated, pointer_updated = lax.cond(
            is_leaf,
            process_leaf,
            process_non_leaf,
            operand=(old_stack, new_stack, stack_pointer),
        )

        return (
            recipient_tree_expression_updated,
            index_array_updated,
            index_pointer_updated,
            old_stack_updated,
            new_stack_updated,
            pointer_updated,
        )

    def condition(state: tuple) -> ScalarBool:
        """Check if the traversal is complete."""
        _, _, _, _, _, stack_pointer = state
        return stack_pointer >= 0

    # iterate over the sample using lax.while_loop
    final_tree_expression, final_index_array, final_index_pointer, _, _, _ = (
        lax.while_loop(
            condition,
            traverse,
            initial_state,
        )
    )
    return final_tree_expression, final_index_array, final_index_pointer



import beartype.typing as tp
import blackjax
import jax
import jax.numpy as jnp
import jax.random as jr
from blackjax.mcmc.hmc import HMCState
from flax import nnx
from jax import jit
from jaxtyping import PRNGKeyArray
from tensorflow_probability.substrates.jax.distributions import Distribution

from gallifrey.data import Dataset
from gallifrey.kernels.prior import KernelPrior
from gallifrey.parameter import ParticleParameter
from gallifrey.particles import transform_particle_parameters
from gallifrey.utils.probability_calculations import calculate_marginal_log_likelihood
from gallifrey.utils.typing import ScalarFloat, ScalarInt


def get_hmc_objective(
    particle_state: nnx.State,
    data: Dataset,
    kernel_prior: KernelPrior,
    noise_prior: Distribution,
) -> tp.Callable:
    """
    Wrapper function to create the HMC objective function.

    The HMC objective function is the sum of the marginal log likelihood
    and the log prior of the kernel and noise parameters.

    The wrapper splits the particle state into a parameter state (
    used for sampling) and a static state.
    The resulting objective function takes the particle parameter
    state in the unconstrained space, transforms it to the constrained
    space, and returns the value of the objective function.

    Parameters
    ----------
    particle_state : nnx.State
        The particle state.
    data : Dataset
        A dataset object containing the input x and output y.
    kernel_prior : KernelPrior
        The kernel prior object.
    noise_prior : Distribution
        The noise prior distribution.

    Returns
    -------
    tp.Callable
        The HMC objective function, which takes the unconstrained
        particle parameter state and returns the value of the objective
        function.
    """

    _, *particle_static_state = particle_state.split(ParticleParameter, ...)

    @jit
    def hmc_objective(
        unconstrained_particle_parameter_state: nnx.State,
    ) -> ScalarFloat:
        """HMC objective function"""

        unconstrained_particle_state = nnx.State.merge(
            unconstrained_particle_parameter_state,
            *particle_static_state,
        )

        constrained_particle_state = transform_particle_parameters(
            unconstrained_particle_state,
            kernel_prior,
            inverse=False,
        )

        kernel_state = constrained_particle_state.kernel
        kernel = kernel_prior.reconstruct_kernel(kernel_state)

        # calcuate marginal log likelihood
        kernel_gram = kernel._gram_train(data.x)
        noise_variance = jnp.array(
            constrained_particle_state.noise_variance.value,
        )  # type: ignore
        mll = calculate_marginal_log_likelihood(
            kernel_gram,
            noise_variance,
            data,
        )

        # calculate priors
        log_prob_kernel_parameter_prior = kernel_prior.parameter_prior.log_prob(
            kernel_state
        )
        log_prob_noise_parameter_prior = jnp.asarray(
            noise_prior.log_prob(noise_variance),
            dtype=data.x.dtype,
        ).squeeze()

        return mll + log_prob_kernel_parameter_prior + log_prob_noise_parameter_prior

    return hmc_objective


def hmc_inference_loop(
    key: PRNGKeyArray,
    hmc_kernel: tp.Callable,
    initial_state: HMCState,
    n_hmc: ScalarInt,
    transform: tp.Callable = lambda state, info: (state, info),
) -> tuple[HMCState, tp.Any]:
    """
    Run an HMC inference loop.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    hmc_kernel : tp.Callable
        The HMC kernel (e.g. HMC, NUTS), derived
        from one of blackjax's HMC sampling algorithms.
    initial_state : HMCState
        The initial state of the HMC sampler.
    n_hmc : ScalarInt
        The number of HMC steps.
    transform : tp.Callable, optional
        A transformation of the trace of states (and info) to be returned.
        This is useful for computing determinstic variables, or
        returning a subset of the states. By default, the states are
        returned as a tuple (state, info).

    Returns
    -------
    HMCState
        The final state.
    tp.Any
        The history of states. The output type depends on the
        transform function. By default, the states are a tuple of
        the state and the info.
    """

    @jit
    def one_step(state: HMCState, key: PRNGKeyArray) -> tuple[HMCState, tp.Any]:
        new_state, info = hmc_kernel(key, state)
        return new_state, transform(new_state, info)

    keys = jr.split(key, int(n_hmc))
    final_state, history = jax.lax.scan(
        one_step,
        initial_state,
        keys,
    )

    return final_state, history


def create_hmc_sampler_factory(
    hmc_config: dict[str, float],
    num_parameter: int,
) -> tp.Callable:
    """
    Create a Blackjax HMC sampler factory.
    The returned function takes an objective function and returns
    a Blackjax HMC sampler.

    This wrapper is used to apply the HMC config to the HMC sampler.

    The inverse mass matrix is diagonal with ones, and scaled by the
    inverse mass matrix scaling factor in the config.

    Parameters
    ----------
    hmc_config : dict[str, float]
        A dictionary containing the HMC config parameters, must
        contain the keys:
        - "step_size"
        - "inv_mass_matrix_scaling"
        - "num_integration_steps"
    num_parameter : int
        The (maximum) number of parameters in the model.

    Returns
    -------
    tp.Callable
        A function that takes an objective function and returns
        a Blackjax HMC sampler.
    """

    def hmc_sampler_factory(
        objective_function: tp.Callable,
    ) -> blackjax.SamplingAlgorithm:

        step_size = hmc_config["step_size"]
        inv_mass_matrix = hmc_config["inv_mass_matrix_scaling"] * jnp.diag(
            jnp.ones(num_parameter)
        )
        num_integration_steps = hmc_config["num_integration_steps"]

        hmc_sampler = blackjax.hmc(
            objective_function,
            step_size,
            inv_mass_matrix,
            num_integration_steps,
        )

        return hmc_sampler

    return hmc_sampler_factory



from dataclasses import field

import jax.numpy as jnp
import tensorflow_probability.substrates.jax.bijectors as tfb
from flax import struct
from jaxtyping import Float
from tensorflow_probability.substrates.jax.distributions import (
    Distribution,
    InverseGamma,
)

from gallifrey.kernels.atoms import (  # noqa: F401
    AbstractAtom,
    AbstractOperator,
    ConstantAtom,
    LinearAtom,
    LinearWithShiftAtom,
    Matern12Atom,
    Matern32Atom,
    Matern52Atom,
    PeriodicAtom,
    PoweredExponentialAtom,
    ProductOperator,
    RationalQuadraticAtom,
    RBFAtom,
    SumOperator,
    WhiteAtom,
)


@struct.dataclass
class GPConfig:
    """
    Config for the GP model.

    Attributes
    ----------

    max_depth : int
        Maximum depth of the kernel tree. By default, 3.

    atoms : list[AbstractAtom]
        List of atomic kernels to consider in the kernel tree.
        By default, the following kernels are included:
        - Linear
        - Periodic
        - RBF

    operators : list[AbstractOperator]
        List of kernel operators to consider in the kernel tree.
        The operators are used to combine the atomic kernels (i.e., functions)
        in the kernel structure. By default, the following operators are included:
        - SumOperator
        - ProductOperator

    node_probabilities :  Float[jnp.ndarray, " D"]
        Probabilities for sampling the kernels and operators. This array
        should have a length equal to the sum of the number of kernels
        and operators. The first part of the array should contain the
        probabilities for sampling the kernels (in the order of them being listed
        in the `atoms` attribute), and the second part should contain the
        probabilities for sampling the operators (in the order of them being listed
        in the `operators` attribute).
        By default, the probabilities are set to be equal for all kernels, and
        half that probability for the operators (to encourage kernels with
        fewer terms)

    prior_transforms : dict[str, Transformation]
        Dictionary containing the bijectors for transforming the kernel parameters
        to the prior distribution. Originally, the parameters are sampled from a
        standard normal distribution, and these bijectors transform the samples to
        the desired prior distribution. The keys are inherited from GPJax, and describe
        the domain of the kernel parameters. The values are the bijectors that
        transform the standard normal samples to the desired prior distribution. The
        bijection transformation should be implemented via TensorFlow Probability
        bijectors.
        By default, the following bijectors are included:
        - "real": Log-normal transform (normal -> log-normal)
        - "positive": Log-normal transform (normal -> log-normal)
        - "sigmoid": Logit-normal transform (normal -> logit-normal)
        NOTE: The "none" key is reserved for internal use and should not be used.

    hmc_config : dict[str, float]
        Configuration for the HMC sampler. The dictionary should contain the following
        keys:
        - "step_size": The step size of the HMC sampler. By default, 0.02.
        - "inv_mass_matrix_scaling": The scaling factor for the inverse mass matrix.
            By default, 1.0.
        - "num_integration_steps": The number of integration steps for the HMC sampler.
            By default, 10.

    noise_prior : Distribution
        An instance of tensorflow_probability.substrates.jax.distributions.Distribution
        that samples the noise variance from the prior distribution. Must have methods
        "sample" and "log_prob". The default is an InverseGamma(1,1) distribution.
        NOTE: The noise variance prior is currently fixed to the InverseGamma(1,1)
        distribution, and cannot be changed. This is because the InverseGamma(1,1)
        distribution is needed to caluculate the Monte Carlo acceptance probabilities
        analytically. Other distributions are not currently supported.

    mean_function : AbstractMeanFunction
        The mean function of the GP model. By default, a zero mean function is used.
        The constant is explicitly set to 0.0, (and is not a ParticleParameter from
        the gallifrey.parameter class, so it is not trainable).
        NOTE: Non-zero mean functions are not currently implemented, do not change
        this attribute.

    """

    max_depth: int = 3

    atoms: list[AbstractAtom] = field(
        default_factory=lambda: [
            # ConstantAtom(),
            LinearAtom(),
            # LinearWithShiftAtom(),
            PeriodicAtom(),
            # Matern12Atom(),
            # Matern32Atom(),
            # Matern52Atom(),
            RBFAtom(),
            # PoweredExponentialAtom(),
            # RationalQuadraticAtom(),
            # WhiteAtom(),
        ]
    )

    operators: list[AbstractOperator] = field(
        default_factory=lambda: [
            SumOperator(),
            ProductOperator(),
        ]
    )

    node_probabilities: Float[jnp.ndarray, " D"] = field(
        default_factory=lambda: jnp.array(
            [
                # 1.0,  # Constant
                1.0,  # Linear
                # 1.0, # LinearWithShift
                1.0,  # Periodic
                # 1.0, # Matern12
                # 1.0, # Matern32
                # 1.0, # Matern52
                1.0,  # RBF
                # 1.0, # PoweredExponential
                # 1.0, # RationalQuadratic
                # 1.0, # White
                0.5,  # SumOperator
                0.5,  # ProductOperator
            ]
        )
    )

    prior_transforms: dict[str, tfb.Bijector] = field(
        default_factory=lambda: dict(
            {
                # this is the transformation y = exp(mu + sigma * z),
                # with mu = 0 and sigma = 1,
                # if z ~ normal(0, 1) then y ~ log-normal(mu, sigma)
                "real": tfb.Chain(
                    [
                        tfb.Exp(),
                        tfb.Shift(jnp.array(0.0)),
                        tfb.Scale(jnp.array(1.0)),
                    ]
                ),
                "positive": tfb.Chain(
                    [
                        tfb.Exp(),
                        tfb.Shift(jnp.array(0.0)),
                        tfb.Scale(jnp.array(1.0)),
                    ]
                ),
                # this is the transformation y = 1/(1 + exp(-(mu + sigma * z))),
                # with mu = 0 and sigma = 1,
                # if z ~ normal(0, 1) then y ~ logit-normal(mu, sigma)
                "sigmoid": tfb.Chain(
                    [
                        tfb.Sigmoid(
                            low=jnp.array(0.0),
                            high=jnp.array(
                                0.95  # slightly below 1 to avoid numerical issues
                            ),
                        ),
                        tfb.Shift(jnp.array(0.0)),
                        tfb.Scale(jnp.array(1.0)),
                    ]
                ),
            }
        )
    )

    hmc_config: dict[str, float] = field(
        default_factory=lambda: {
            "step_size": 0.02,
            "inv_mass_matrix_scaling": 1.0,
            "num_integration_steps": 10,
        }
    )

    @property
    def noise_prior(self) -> Distribution:
        """
        The distribution of the noise variance prior,
        specifically an InverseGamma(1,1) distribution.

        Returns
        -------
        Distribution
            The noise variance prior distribution.
        """
        return InverseGamma(jnp.array(1.0), jnp.array(1.0))



import os

from beartype import BeartypeConf
from beartype.claw import beartype_this_package
from jax import config as jax_config

# type checks the entire package, but only throw warnings
# (must be before gallifrey imports)
beartype_this_package(conf=BeartypeConf(violation_type=UserWarning))  # type: ignore

from gallifrey.model import GPConfig, GPModel  # noqa: E402
from gallifrey.schedule import LinearSchedule, LogSchedule  # noqa: E402

os.environ["JAX_ENABLE_X64"] = "True"
jax_config.update("jax_enable_x64", True)
# no idea why, but this increases performance by a factor of 2, at least on CPU
os.environ["OMP_NUM_THREADS"] = "1"

print("gallifrey: Setting flag `JAX_ENABLE_X64` to `True`")
print("gallifrey: Setting flag `OMP_NUM_THREADS` to `1`")


__all__ = [
    "GPConfig",
    "GPModel",
    "LinearSchedule",
    "LogSchedule",
]



from abc import ABC, abstractmethod

import jax.numpy as jnp


class Schedule(ABC):
    """
    Abstract base class for SMC data annealing schedules.

    """

    @staticmethod
    @abstractmethod
    def generate(
        num_datapoints: int,
        num_steps: int,
        start: int = 1,
    ) -> tuple[int, ...]:
        """
        Generate an annealing schedule.

        Parameters
        ----------
        num_datapoints : int
            The total number of datapoints.
        num_steps : int
            The number of steps in the schedule.
        start : int, optional
            The starting point of the schedule,
            by default 1 (one observation).

        Returns
        -------
        tuple[int, ...]
            A tuple of integers representing the cumulative number of
            observations at each step of the schedule. Length of the
            tuple is `num_steps`.

        """
        pass


class LinearSchedule(Schedule):
    """
    Linear scheduler, adds roughly
    `n * percent` new observations at each step.

    """

    @staticmethod
    def generate(
        num_datapoints: int,
        num_steps: int,
        start: int = 1,
    ) -> tuple[int, ...]:
        """
        Generate a linear annealing schedule.

        Parameters
        ----------
        num_datapoints : int
            The total number of datapoints.
        num_steps : int
            The number of steps in the schedule.
        start : int, optional
            The starting point of the schedule,
            by default 1 (one observation).

        Returns
        -------
        tuple[int, ...]
            A tuple of integers representing the cumulative number of
            observations at each step of the schedule. Length of the
            tuple is `num_steps`.

        """
        # if only one step, run total number of datapoints
        if num_steps == 1:
            start = num_datapoints

        return tuple(
            jnp.round(
                jnp.linspace(
                    start=start,
                    stop=num_datapoints,
                    num=num_steps,
                    endpoint=True,
                    dtype=float,
                ),
            )
            .astype(int)
            .tolist()
        )


class LogSchedule(Schedule):
    """
    A logarithmic scheduler, adds observations in a logarithmic fashion.

    """

    @staticmethod
    def generate(
        num_datapoints: int,
        num_steps: int,
        start: int = 1,
    ) -> tuple[int, ...]:
        """
        Generate a logarithmic annealing schedule.

        Parameters
        ----------
        num_datapoints : int
            The total number of datapoints.
        num_steps : int
            The number of steps in the schedule.
        start : int, optional
            The starting point of the schedule,
            by default 1 (one observation).

        Returns
        -------
        tuple[int, ...]
            A tuple of integers representing the cumulative number of
            observations at each step of the schedule. Length of the
            tuple is `num_steps`.
        """
        # if only one step, run total number of datapoints
        if num_steps == 1:
            start = num_datapoints

        return tuple(
            jnp.round(
                jnp.geomspace(
                    start=start,
                    stop=num_datapoints,
                    num=num_steps,
                    endpoint=True,
                )
            )
            .astype(int)
            .tolist()
        )



from functools import partial

import jax.numpy as jnp
from flax import nnx
from jax import debug, jit, lax
from tensorflow_probability.substrates.jax.distributions import Distribution

from gallifrey.data import Dataset
from gallifrey.kernels.prior import KernelPrior
from gallifrey.moves.noise_proposal import noise_variance_probability
from gallifrey.utils.probability_calculations import calculate_marginal_log_likelihood
from gallifrey.utils.typing import ScalarFloat


@jit
def calculate_acceptance_probability(
    hastings_ratio_numerator: ScalarFloat,
    hastings_ratio_denominator: ScalarFloat,
) -> ScalarFloat:
    """
    Calculate the acceptance ratio for a given pair of
    hastings ratio terms.
    In the context of the particle move, the numerator
    is associated with the proposed state, and the denominator
    with the current state.

    Parameters
    ----------
    hastings_ratio_numerator : ScalarFloat
        The hastings ratio contribution for the proposed state.
    hastings_ratio_denominator : ScalarFloat
        The hastings ratio contribution for the current state.
    Returns
    -------
    ScalarFloat
        The MCMC acceptance probability.
    """

    # dealing with NaNs in such a way that the kernel with NaN is rejected,
    # if both are NaN (shouldn't happen), accept proposal and hope for the best
    # NOTE: in general, NaNs should not occur, but they can occur in the
    # kernel gram matrix if the kernel is not positive definite (e.g. due to
    # numerical instabilities)

    log_hastings_ratio = jnp.nan_to_num(
        hastings_ratio_numerator, nan=-jnp.inf
    ) - jnp.nan_to_num(hastings_ratio_denominator, nan=-jnp.inf)
    log_hastings_ratio = jnp.nan_to_num(log_hastings_ratio, nan=0.0)

    acceptance_probability = jnp.clip(
        jnp.exp(log_hastings_ratio),
        max=1.0,
    )

    return acceptance_probability


@partial(
    jit,
    static_argnames=(
        "kernel_prior",
        "data",
        "fix_noise",
        "verbosity",
    ),
)
def calculate_acceptance_prob_contribution(
    particle_state: nnx.State,
    proposed_noise_variance: ScalarFloat,
    kernel_prior: KernelPrior,
    noise_prior: Distribution,
    data: Dataset,
    fix_noise: bool,
    verbosity: int = 0,
) -> ScalarFloat:
    """
    Calculate the contribution to the acceptance probability
    for a specific particle state and noise variance proposal.

    Parameters
    ----------
    particle_state : nnx.State
        The state of the particle.
    proposed_noise_variance : ScalarFloat
        The noise variance of the proposed state. (which is not
        the current state!)
    kernel_prior : KernelPrior
        The kernel prior.
    noise_prior : Distribution
        The prior distribution of the noise variance.
    data : Dataset
        The observational data.
    fix_noise : bool
        Whether to fix the noise variance is fixed or not.
    verbosity : int, optional
        The verbosity level, by default 0. Debugging information
        is printed if `verbosity > 1`.

    Returns
    -------
    ScalarFloat
        The contribution to the acceptance probability for the
        given particle state and noise variance proposal.

    """

    # unpack particle state
    kernel_state = particle_state.kernel
    current_noise_variance = jnp.asarray(particle_state.noise_variance.value)

    kernel = kernel_prior.reconstruct_kernel(kernel_state)

    # calculate gram matrix for kernel
    kernel_gram = kernel._gram_train(data.x)

    # calculate acceptance probability terms
    log_tree_size = jnp.log(kernel.node_sizes[0])
    marginal_log_likelihood = calculate_marginal_log_likelihood(
        kernel_gram,
        current_noise_variance,
        data,
    )
    log_prob_noise_prior = lax.cond(
        fix_noise,
        lambda x: jnp.array(0.0),
        lambda x: jnp.asarray(
            noise_prior.log_prob(x),
            dtype=data.x.dtype,
        ).squeeze(),
        current_noise_variance,
    )

    # calculate noise posterior probability
    # NOTE: We have to calculate the probability of the current
    # noise variance coming from the proposed noise variance,
    # so the two are swapped in the function call.
    # (eq 21 and 22 in Saad2023)
    log_prob_noise = lax.cond(
        fix_noise,
        lambda: jnp.array(0.0),
        lambda: noise_variance_probability(
            kernel_gram=kernel_gram,
            current_noise_variance=proposed_noise_variance,
            proposed_noise_variance=current_noise_variance,
            data=data,
        ),
    )
    # calculate acceptance probability contribution,
    # (eq 22. and Preposition 2 in Saad2023)
    log_acceptance_prob_contribution = jnp.array(
        [
            marginal_log_likelihood,
            -log_tree_size,
            log_prob_noise_prior,
            -log_prob_noise,
        ]
    ).sum()

    if verbosity > 1:
        debug.print("Current Terms:")
        debug.print("Marginal log likelihood: {}", marginal_log_likelihood)
        debug.print("Log (tree size^-1): {}", -log_tree_size)
        debug.print("Log noise prior probability: {}", log_prob_noise_prior)
        debug.print("Log (probability of noise posterior^-1): {}", -log_prob_noise)
        debug.print("Sum of terms: {}", log_acceptance_prob_contribution)
        debug.print("-" * 50)

    return log_acceptance_prob_contribution



#########################
# Code taken from GPJax
#########################

# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import TypeAlias, Union

from jaxtyping import Array as JAXArray
from jaxtyping import Bool, Float, Int
from numpy import ndarray as NumpyArray

Array: TypeAlias = Union[JAXArray, NumpyArray]

ScalarArray: TypeAlias = Float[Array, ""]
ScalarBool: TypeAlias = Union[bool, Bool[Array, ""]]
ScalarInt: TypeAlias = Union[int, Int[Array, ""]]
ScalarFloat: TypeAlias = Union[float, Float[Array, ""]]



from copy import deepcopy
from functools import partial

import beartype.typing as tp
import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from blackjax.smc.resampling import stratified
from flax import nnx
from jax import jit, lax, pmap, vmap
from jax.scipy.special import logsumexp
from jaxtyping import Float, PRNGKeyArray
from tensorflow_probability.substrates.jax.distributions import Distribution

from gallifrey.data import Dataset
from gallifrey.inference.rejuvenation import rejuvenate_particle
from gallifrey.inference.state import GPState
from gallifrey.kernels.prior import KernelPrior
from gallifrey.utils.probability_calculations import calculate_marginal_log_likelihood
from gallifrey.utils.typing import ScalarFloat, ScalarInt


def calculate_mll_wrapper(
    particle_state: nnx.State,
    data: Dataset,
    kernel_prior: KernelPrior,
) -> ScalarFloat:
    """
    Wapper around calculate_marginal_log_likelihood, which
    calculates the MLL from a particle state.

    Parameters
    ----------
    particle_state : nnx.State
        The particle state.
    data : Dataset
        The dataset object, containing the input x and output y.
    kernel_prior : KernelPrior
        The kernel prior, used to reconstruct the kernel.

    Returns
    -------
    ScalarFloat
        The marginal log likelihood for the given particle state and
        subset of the data.

    """

    kernel = kernel_prior.reconstruct_kernel(particle_state.kernel)
    kernel_gram = kernel._gram_train(data.x)

    noise_variance = jnp.array(particle_state.noise_variance.value)  # type: ignore

    mll = calculate_marginal_log_likelihood(
        kernel_gram,
        noise_variance,
        data,
    )
    return mll


@partial(jit, static_argnames=("data", "kernel_prior"))
def reweight_particles(
    smc_state: GPState,
    data: Dataset,
    kernel_prior: KernelPrior,
) -> Float[jnp.ndarray, " N"]:
    """
    Reweights particles based on the marginal log likelihood
    calculated with new data points.

    This function updates the log weights of the particles in the SMC state by
    calculating the marginal log likelihood for each particle using the new subset
    vs the previous subset, taking the difference in MLL for each particle and adding
    it to the existing log weights, in essence:

        w_t = w_{t-1} * (mll_t / mll_{t-1}) (but in log space, so addition instead)

    See Saad2023, algorithm 1 line 5.

    The weights are returned normalised.

    Parameters
    ----------
    smc_state : GPState
        The SMC state after the previous round.
    data : Dataset
        The dataset containing the data at the current time point.
    kernel_prior : KernelPrior
        The kernel prior object, used to construct the kernel.

    Returns
    -------
    Float[jnp.ndarray, " N"]
        The updated (normalised) log weights for the particles.
    """

    new_mlls = vmap(calculate_mll_wrapper, in_axes=(0, None, None))(
        smc_state.particle_states,  # type: ignore
        data,
        kernel_prior,
    )

    mll_ratio = new_mlls - smc_state.marginal_log_likelihoods  # type: ignore

    unnorm_log_weights: jnp.ndarray = smc_state.log_weights + mll_ratio  # type: ignore
    log_weights = unnorm_log_weights - logsumexp(unnorm_log_weights)
    return log_weights


def calculate_effective_sample_size(
    log_weights: Float[jnp.ndarray, " N"]
) -> ScalarFloat:
    """
    Calculates the effective sample size (ESS) of the particles.

    The effective sample size is a measure of how diverse the particle weights are,
    and is used to determine whether resampling is needed. If the ESS is low, it means
    that many particles have similar weights, and resampling is needed to avoid
    degeneracy.

    The ESS is calculated as:

        ESS = 1 / sum(w^2)

    where w is the normalised(!) weight of each particle.

    Parameters
    ----------
    log_weights : Float[jnp.ndarray, " N"]
        The log weights of the particles.

    Returns
    -------
    ScalarFloat
        The effective sample size of the particles.

    """

    weights = jnp.exp(log_weights)
    ess = 1 / jnp.sum(weights**2)
    return ess


@partial(jit, static_argnames=("num_particles", "resampling_func"))
def resample_particles(
    key: PRNGKeyArray,
    particle_states: nnx.State,
    log_weights: Float[jnp.ndarray, " N"],
    num_particles: ScalarInt,
    resampling_func: tp.Callable,
) -> tuple[nnx.State, Float[jnp.ndarray, " N"]]:
    """
    Resamples particles based on the log weights.

    This function resamples the particles based on the log weights
    using a resampling function, which takes the log weights and
    the number of particles and returns the indices of the resampled
    particles.

    The resampling function must have the signature:
        resampling_func(key, weights, num_particles) -> jnp.ndarray

    Note that the actual weights must be passed to the resampling function
    (not the log weights), and that they need to be normalised.

    We use sampling functions from blackjax. By default we use stratified
    resampling, but systematic, multinomial, and residual resampling are also
    available.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key for the resampling function.
    particle_states : nnx.State
        The current particle states.
    log_weights : Float[jnp.ndarray, " N"]
        The log weights of the particles.
    num_particles : ScalarInt
        The number of particles (should be the same as the length of the log weights,
        and the number of particles in the particle_states).
    resampling_func : tp.Callable
        The resampling function to use.

    Returns
    -------
    nnx.State
        The resampled particle states.
    Float[jnp.ndarray, " N"]
        The normalised log weights of the resampled particles.
        (i.e. log(1/N) for each particle, see Saad2023, algorithm 1 line 9).
    """

    # sample indices based on weights
    sampled_indices = resampling_func(key, jnp.exp(log_weights), num_particles)

    # update particle states based on sampled indices
    resampled_state = jtu.tree_map(lambda leaf: leaf[sampled_indices], particle_states)

    resampled_log_weights = jnp.log(jnp.ones(num_particles) / num_particles)

    return resampled_state, resampled_log_weights


def smc_round(
    smc_state: GPState,
    num_particles: ScalarInt,
    n_mcmc: ScalarInt,
    n_hmc: ScalarInt,
    data_subset: Dataset,
    kernel_prior: KernelPrior,
    noise_prior: Distribution,
    fix_noise: bool,
    hmc_sampler_factory: tp.Callable,
    resampling_func: tp.Callable,
    ess_threshold: float,
    total_data_points: ScalarInt,
    verbosity: ScalarInt = 0,
) -> GPState:
    """
    Performs one round of Sequential Monte Carlo (SMC).

    This function executes a single SMC round, which includes reweighting particles
    based on new data, resampling particles if the effective sample size is below
    a threshold, and rejuvenating particles using MCMC and HMC steps.

    Parameters
    ----------
    smc_state : GPState
        The current SMC state.
    num_particles : ScalarInt
        The number of particles.
    n_mcmc : ScalarInt
        The number of MCMC steps to perform during rejuvenation.
    n_hmc : ScalarInt
        The number of HMC steps within each MCMC step during rejuvenation.
    data_subset : Dataset
        The data to be considered for this SMC round.
    kernel_prior : KernelPrior
        The kernel prior object.
    noise_prior : Distribution
        The prior distribution for the noise parameter.
    fix_noise : bool
        Whether to fix the noise parameter during rejuvenation.
    hmc_sampler_factory : tp.Callable
        A factory function that creates an HMC sampler.
    resampling_func : tp.Callable
        The resampling function to use.
    ess_threshold : float
        The threshold for the effective sample size (ESS) below which
        resampling is performed.
    total_data_points : ScalarInt
        The total number of data points in the dataset.
    verbosity : ScalarInt, optional
        The verbosity level for the rejuvenation process. Debug information
        is printed if verbosity > 0.

    Returns
    -------
    SMState
        The updated SMC state after the round.

    """

    key, resample_key, rejuvenate_key, state_key = jr.split(
        smc_state.key,  # type: ignore
        4,
    )

    # reweight
    new_log_weights = reweight_particles(
        smc_state,
        data_subset,
        kernel_prior,
    )
    if verbosity > 0:
        print(f"Weights: {jnp.exp(new_log_weights)}")

    # get number of data points for this round
    num_data_points = data_subset.x.shape[0]

    # maybe resample, depending on effective sample size (ESS)
    # (if resample, also reset log weights to uniform)
    normalised_ess = (
        calculate_effective_sample_size(new_log_weights)
        / smc_state.num_particles  # type: ignore
    )
    # do not resample if we are at last step
    do_resample = (normalised_ess < ess_threshold) & (
        num_data_points < total_data_points
    )

    resampled_particle_states, new_log_weights, resampled = lax.cond(
        do_resample,
        lambda key: (
            *resample_particles(
                resample_key,
                smc_state.particle_states,
                new_log_weights,
                num_particles,
                resampling_func,
            ),
            True,
        ),
        lambda key: (smc_state.particle_states, new_log_weights, False),
        key,
    )
    if verbosity > 0:
        print(
            f"Resampled: {resampled} (Normalised ESS: "
            f"{float(normalised_ess):.2f})"  # type: ignore
        )

    # rejuvenate
    @jit
    def wrapper(
        key: PRNGKeyArray,
        state: nnx.State,
    ) -> tuple[nnx.State, nnx.State, ScalarInt, ScalarInt]:
        """Wrapper around the rejuvenate_particle function using
        GPmodel attributes."""
        return rejuvenate_particle(
            key,
            state,
            data_subset,
            kernel_prior,
            noise_prior,
            n_mcmc=n_mcmc,
            n_hmc=n_hmc,
            fix_noise=fix_noise,
            hmc_sampler_factory=hmc_sampler_factory,
            verbosity=verbosity,
        )

    # run rejuvenation on all particles, parallelised
    final_state, _, accepted_mcmc, accepted_hmc = pmap(wrapper, in_axes=0)(
        jr.split(rejuvenate_key, num_particles),  # type: ignore
        resampled_particle_states,
    )

    if verbosity > 0:
        for i, acc_mcmc, acc_hmc in zip(
            range(smc_state.num_particles),  # type: ignore
            accepted_mcmc,
            accepted_hmc,
        ):
            print(
                f"Particle {i+1} | Accepted: MCMC[{acc_mcmc}/{n_mcmc}] "
                f" HMC[{acc_hmc}/{acc_mcmc*n_hmc}]"
            )
        print("=" * 50)

    # calculate annealing MLL
    new_mlls = jnp.array(
        jit(
            vmap(calculate_mll_wrapper, in_axes=(0, None, None)),
            static_argnames=("data", "kernel_prior"),
        )(
            final_state,
            data_subset,
            kernel_prior,
        )
    )

    return GPState(
        particle_states=final_state,
        log_weights=new_log_weights,
        marginal_log_likelihoods=new_mlls,
        num_particles=smc_state.num_particles,
        num_data_points=num_data_points,
        resampled=resampled,
        mcmc_accepted=accepted_mcmc,
        hmc_accepted=accepted_hmc,
        key=state_key,
    )


# @partial(
#     jit,
#     static_argnames=(
#         "annealing_schedule",
#         "num_particles",
#         "data",
#         "kernel_prior",
#         "noise_prior",
#         "fix_noise",
#         "hmc_sampler_factory",
#         "n_mcmc",
#         "n_hmc",
#         "resampling_func",
#         "verbosity",
#     ),
# )
def smc_loop(
    key: PRNGKeyArray,
    particle_states: nnx.State,
    annealing_schedule: tuple[int, ...],
    num_particles: int,
    data: Dataset,
    kernel_prior: KernelPrior,
    noise_prior: Distribution,
    fix_noise: bool,
    hmc_sampler_factory: tp.Callable,
    n_mcmc: ScalarInt,
    n_hmc: ScalarInt,
    resampling_func: tp.Callable = stratified,
    ess_threshold: float = 0.5,
    verbosity: ScalarInt = 0,
) -> tuple[GPState, list[GPState]]:
    """
    Runs the Sequential Monte Carlo (SMC) algorithm.

    The SMC loop runs a series of SMC rounds using a data annealing
    schedule, each round includes three steps:

    - Reweighting: Update particle weights based on new data.
    The new weight is the old weight times the ratio of the new
    marginal likelihood evaluated on the new data (so all data up
    to n_t, where n_t is the number of data points at time t in
    the annealing schedule) and the old marginal likelihood evaluated
    on the previous data (up to n_{t-1}). In essence, the new weight
    reflect how much better the particle fits the new (unseen) data
    compared to the data it was trained on.

    - Resampling: Adaptively resample particles based on the new weights.
    This step is performed if the effective sample size (ESS) is below a
    threshold (0.5 in this implementation). The ESS is a measure of how
    diverse the particle weights are, and if it is low, it means that many
    particles have similar weights, and resampling is needed to avoid
    degeneracy. There are various resampling strategies available, we
    default to stratified resampling.

    - Rejuvenation: Rejuvenate particles using MCMC and HMC steps. This
    step is performed on all particles, and consists of two main moves:
        1. Structure Move: Proposes a new kernel structure based on the
        `kernel_prior`.
        2. Parameter Rejuvenation: If the structure move is accepted,
        rejuvenates the kernel parameters and potentially the noise variance
        using Hamiltonian Monte Carlo (HMC).



    Parameters
    ----------
    key : PRNGKeyArray
        The random key used for sampling.
    particle_states : nnx.State
        The particle states at the start of the SMC loop.
    annealing_schedule : Int[jnp.ndarray, " T"]
        The annealing schedule, an array of integers representing
        the number of data points to use at each time point
        (e.g. jnp.array([10, 20, 30, 40, 50]) ).
    num_particles : int
        The number of particles (must match number of particles
        in the particle_states).
    data : Dataset
        The dataset object containing the input x and output y for
        all data points.
    kernel_prior : KernelPrior
        The kernel prior object, used to reconstruct the kernel, and sample
        new kernel structures.
    noise_prior : Distribution
        The prior distribution for the noise variance.
    fix_noise : bool
        Whether to fix the noise variance during rejuvenation, or sample it.
    hmc_sampler_factory : tp.Callable
        A factory function that creates an HMC sampler instance for the parameter
        rejuvenation step.
    n_mcmc : ScalarInt
        The number of MCMC steps to perform during rejuvenation.
    n_hmc : ScalarInt
        The number of HMC steps to perform during rejuvenation.
    resampling_func : tp.Callable, optional
        The resampling function to use, by default stratified.
    ess_threshold : float, optional
        The threshold for the effective sample size (ESS) below which
        resampling is performed, by default 0.5.
    verbosity : int, optional
        The verbosity level for the rejuvenation process. Debug information
        is printed if verbosity > 0.

    Returns
    -------
    GPState
        The final SMC state after the SMC loop.
    list[GPState]
        The full history of the SMC state at each time point in the
        annealing schedule.

    """

    def smc_round_wrapper(
        smc_state: GPState,
        points_to_use: int,
    ) -> GPState:
        """Wrapper around the smc_round function."""

        x_slice = lax.slice(data.x, (0,), (points_to_use,))
        y_slice = lax.slice(data.y, (0,), (points_to_use,))

        kernel_prior_round = deepcopy(kernel_prior)
        kernel_prior_round.num_datapoints = points_to_use

        data_subset = Dataset(x=x_slice, y=y_slice)

        return smc_round(
            smc_state,
            num_particles,
            n_mcmc,
            n_hmc,
            data_subset,
            kernel_prior_round,
            noise_prior,
            fix_noise,
            hmc_sampler_factory,
            resampling_func,
            ess_threshold,
            data.x.shape[0],
            verbosity,
        )

    # initialise SMC state
    initial_smc_state = GPState(
        particle_states=particle_states,
        log_weights=jnp.log(jnp.ones(num_particles) / num_particles),
        marginal_log_likelihoods=jnp.zeros(num_particles),
        num_particles=num_particles,
        num_data_points=jnp.array(0),
        resampled=False,
        mcmc_accepted=jnp.zeros(num_particles, dtype=int),
        hmc_accepted=jnp.zeros(num_particles, dtype=int),
        key=key,
    )

    history = []
    for i, points_to_use in enumerate(annealing_schedule):
        if verbosity > 0:
            print(
                f"Running SMC round [{i+1}/{len(annealing_schedule)}] "
                f"with [{points_to_use}/{len(data.x)}] data points."
            )

        final_smc_state = smc_round_wrapper(
            initial_smc_state,
            points_to_use,
        )
        initial_smc_state = final_smc_state
        history.append(final_smc_state)

    return final_smc_state, history



import jax.numpy as jnp
from jaxtyping import Float, PRNGKeyArray
from tensorflow_probability.substrates.jax.distributions import InverseGamma

from gallifrey.data import Dataset
from gallifrey.utils.probability_calculations import calculate_inv_sigma_r
from gallifrey.utils.typing import ScalarFloat


def noise_variance_proposal(
    key: PRNGKeyArray,
    kernel_gram: Float[jnp.ndarray, "D D"],
    noise_variance: ScalarFloat,
    data: Dataset,
) -> tuple[ScalarFloat, ScalarFloat]:
    """
    Sample a new proposal for the noise variance,
    and the log probability of the proposal.

    See equation 21 in Saad et al. 2023.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    kernel_gram : Float[jnp.ndarray, "D D"]
        The gram matrix of the kernel.
    noise_variance : ScalarFloat
        The current noise variance.
    data : Dataset
        The data containing the observations,
        as a Dataset object.

    Returns
    -------
    ScalarFloat
        The new noise variance.
    ScalarFloat
        The log probability of the proposal.
    """

    # calculate Σ⁻¹y | [n,1]
    inv_sigma_r = calculate_inv_sigma_r(
        data.y,
        kernel_gram,
        noise_variance,
    )

    # calculate the means of the Gaussian process, µ = Kxx*Σ⁻¹y | [n,1]
    means = kernel_gram @ inv_sigma_r

    # sample new noise variance from inverse gamma distribution
    diff = data.y - means
    inv_gamma = InverseGamma(
        concentration=1 + data.y.size / 2,
        scale=1 + 0.5 * diff @ diff,
    )
    new_noise_variance = inv_gamma.sample(seed=key, sample_shape=())
    log_prob = inv_gamma.log_prob(new_noise_variance)

    return new_noise_variance, log_prob  # type: ignore


def noise_variance_probability(
    kernel_gram: Float[jnp.ndarray, "D D"],
    current_noise_variance: ScalarFloat,
    proposed_noise_variance: ScalarFloat,
    data: Dataset,
) -> ScalarFloat:
    """
    Calculate the posterior probability for a
    proposed noise variance, given the current
    kernel gram matrix and current noise variance.

    See equation 21 in Saad2023.

    Parameters
    ----------
    kernel_gram : Float[jnp.ndarray, "D D"]
        The gram matrix of the kernel.
    current_noise_variance : ScalarFloat
        The current noise variance.
    proposed_noise_variance : ScalarFloat
        The proposed noise variance.
    data : Dataset
        The data containing the observations,
        as a Dataset object.

    Returns
    -------
    ScalarFloat
        The posterior probability for the proposed noise variance.
    """

    # calculate Σ⁻¹y | [n,1]
    inv_sigma_r = calculate_inv_sigma_r(
        data.y,
        kernel_gram,
        current_noise_variance,
    )

    # calculate the means of the Gaussian process, µ = Kxx*Σ⁻¹y | [n,1]
    means = kernel_gram @ inv_sigma_r

    # create inverse gamma distribution
    diff = data.y - means
    inv_gamma = InverseGamma(
        concentration=1 + data.y.size / 2,
        scale=1 + 0.5 * diff @ diff,
    )
    # calculate the log probability of the proposal
    log_prob_proposal = inv_gamma.log_prob(proposed_noise_variance)

    return log_prob_proposal



from copy import deepcopy
from functools import partial

import jax.numpy as jnp
import jax.random as jr
from flax import nnx
from jax import debug, jit, lax
from jaxtyping import PRNGKeyArray
from tensorflow_probability.substrates.jax.distributions import Distribution

from gallifrey.data import Dataset
from gallifrey.kernels.prior import KernelPrior
from gallifrey.moves.acceptance_probability import (
    calculate_acceptance_prob_contribution,
    calculate_acceptance_probability,
)
from gallifrey.moves.detach_attach import detach_attach_move
from gallifrey.moves.noise_proposal import noise_variance_proposal
from gallifrey.moves.subtree_replace import subtree_replace_move
from gallifrey.utils.probability_calculations import calculate_marginal_log_likelihood
from gallifrey.utils.typing import ScalarBool, ScalarFloat


@partial(
    jit,
    static_argnames=(
        "kernel_prior",
        "data",
        "fix_noise",
        "verbosity",
    ),
)
def structure_move(
    key: PRNGKeyArray,
    particle_state: nnx.State,
    kernel_prior: KernelPrior,
    noise_prior: Distribution,
    data: Dataset,
    fix_noise: bool,
    verbosity: int = 0,
) -> tuple[nnx.State, ScalarBool]:
    """
    Perform a structure move on the particle state,
    by proposing a new kernel structure (via a detach-attach
    or subtree-replace move), and a new noise variance.

    The acceptance probability is calculated based on the current
    and proposed particle states, and the proposal is accepted
    stochasticly based on that.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key for the move.
    particle_state : nnx.State
        The current particle state.
    kernel_prior : KernelPrior
        The kernel prior used for sampling a new
        kernel structure.
    noise_prior : Distribution
        The noise variance prior, used for calculating
        the contribution of the noise variance to the
        acceptance probability.
    data : Dataset
        The data, a Dataset object containing the
        observations. (The observations y should already
        be centered around the deterministic mean function,
        i.e. it should be the residuals.)
    fix_noise : bool
        Whether to fix the noise variance, or sample it.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information
        is printed if verbosity > 1.

    Returns
    -------
    nnx.State
        The new particle state after the move.
    ScalarBool
        Whether the move was accepted.
    """

    key, proposal_key, acceptance_key = jr.split(key, 3)

    proposed_particle_state, log_hastings_ratio_numerator = particle_state_proposal(
        proposal_key,
        particle_state,
        kernel_prior,
        noise_prior,
        data,
        fix_noise=fix_noise,
        verbosity=verbosity,
    )

    log_hastings_ratio_denominator = calculate_acceptance_prob_contribution(
        particle_state,
        proposed_particle_state.noise_variance.value,  # type: ignore
        kernel_prior,
        noise_prior,
        data,
        fix_noise=fix_noise,
        verbosity=verbosity,
    )

    acceptance_probability = calculate_acceptance_probability(
        log_hastings_ratio_numerator,
        log_hastings_ratio_denominator,
    )

    accepted = jr.bernoulli(acceptance_key, acceptance_probability)
    if verbosity > 1:
        debug.print("-" * 50)
        debug.print("Structure move accepted: {}", accepted)
        debug.print("Acceptance Probability: {}", acceptance_probability)
        debug.print("-" * 50)

    new_particle_state = lax.cond(
        accepted,
        lambda: proposed_particle_state,
        lambda: particle_state,
    )

    return new_particle_state, accepted


@partial(
    jit,
    static_argnames=(
        "kernel_prior",
        "data",
        "fix_noise",
        "verbosity",
    ),
)
def particle_state_proposal(
    key: PRNGKeyArray,
    particle_state: nnx.State,
    kernel_prior: KernelPrior,
    noise_prior: Distribution,
    data: Dataset,
    fix_noise: bool,
    p_detach_attach: ScalarFloat = 0.5,
    verbosity: int = 0,
) -> tuple[nnx.State, ScalarFloat]:
    """
    Propose a new particle state by proposing a new kernel structure
    and noise variance, and calculating the contribution to the
    acceptance probability.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    particle_state : nnx.State
        The current particle state.
    kernel_prior : KernelPrior
        The kernel prior.
    noise_prior : Distribution
        The noise variance prior.
    data : Dataset
        The data.
    fix_noise : bool
        Whether to fix the noise variance. In case
        of fixed noise, the noise variance is not updated.
    p_detach_attach : ScalarFloat, optional
        The probability of performing a detach-attach move
        (vs a subtree-replace move), by default 0.5.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information
        is printed if verbosity > 1.

    Returns
    -------
    nnx.State
        The proposed particle state.
    ScalarFloat
        The log acceptance probability numerator.
    """

    # get necessary keys
    key, structure_key, noise_key = jr.split(key, 3)

    # unpack particle state
    kernel_state = particle_state.kernel
    noise_variance = jnp.array(particle_state.noise_variance.value)  # type: ignore

    # propose new kernel structure
    proposed_kernel_state, detach_attach_log_prob, performed_detach_attach = (
        kernel_structure_proposal(
            structure_key,
            kernel_state,
            kernel_prior,
            p_detach_attach,
            verbosity=verbosity,
        )
    )

    proposed_kernel = kernel_prior.reconstruct_kernel(proposed_kernel_state)

    # calculate gram matrix for new kernel
    kernel_gram = proposed_kernel._gram_train(data.x)

    # propose new noise variance
    proposed_noise_variance, log_prob_noise = lax.cond(
        fix_noise,
        lambda key: (noise_variance, 0.0),  # return noise variance as is
        lambda key: noise_variance_proposal(
            key,
            kernel_gram,
            noise_variance,
            data,
        ),
        noise_key,
    )

    # update particle state
    new_par_state = deepcopy(particle_state)
    new_par_state.kernel = proposed_kernel_state  # type: ignore
    new_par_state.noise_variance = (  # type: ignore
        particle_state.noise_variance.replace(  # type: ignore
            proposed_noise_variance,
        )
    )

    # calculate acceptance probability terms
    log_tree_size = jnp.log(proposed_kernel.node_sizes[0])
    marginal_log_likelihood = calculate_marginal_log_likelihood(
        kernel_gram,
        proposed_noise_variance,
        data,
    )
    log_prob_noise_prior = lax.cond(
        fix_noise,
        lambda x: jnp.array(0.0),
        lambda x: jnp.asarray(
            noise_prior.log_prob(x),
            dtype=data.x.dtype,
        ).squeeze(),
        proposed_noise_variance,
    )

    # calculate acceptance probability contribution,
    # (eq 22. and Preposition 2 in Saad2023)
    log_acceptance_prob_numerator = jnp.array(
        [
            marginal_log_likelihood,
            -log_tree_size,
            log_prob_noise_prior,
            -log_prob_noise,
            detach_attach_log_prob,
        ]
    ).sum()

    if verbosity > 1:
        lax.cond(
            performed_detach_attach,
            lambda: debug.print("Move Type: Detach-attach"),
            lambda: debug.print("Move Type: Subtree-replace"),
        )
        debug.print("-" * 50)
        debug.print("Proposal Terms:")
        debug.print("Marginal log likelihood: {}", marginal_log_likelihood)
        debug.print("Log (tree size^-1): {}", -log_tree_size)
        debug.print("Log noise prior probability: {}", log_prob_noise_prior)
        debug.print("Log (probability of noise posterior^-1): {}", -log_prob_noise)
        debug.print("Log detach-attach probability: {}", detach_attach_log_prob)
        debug.print("Sum of terms: {}", log_acceptance_prob_numerator)
        debug.print("-" * 50)

    return new_par_state, log_acceptance_prob_numerator


@partial(jit, static_argnames=("kernel_prior", "verbosity", "p_detach_attach"))
def kernel_structure_proposal(
    key: PRNGKeyArray,
    kernel_state: nnx.State,
    kernel_prior: KernelPrior,
    p_detach_attach: ScalarFloat = 0.5,
    verbosity: int = 0,
) -> tuple[nnx.State, ScalarFloat, ScalarBool]:
    """
    Perform a move on the kernel structure,
    either detach-attach or subtree-replace.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    kernel_state : nnx.State
        The original kernel state.
    kernel_prior : KernelPrior
        The kernel prior.
    p_detach_attach : ScalarFloat, optional
        The probability of performing a detach-attach move
        (vs a subtree-replace move), by default 0.5.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information
        for the detach-attach and subtree-replace moves
        are printed if verbosity > 2.

    Returns
    -------
    nnx.State
        The proposed kernel state.
    ScalarFloat
        The log probability associated with the detach-attach move
        (0.0 if a subtree-replace move is performed).
    """

    key, detach_attach_key, move_key = jr.split(key, 3)

    # choose move to perform (detach-attach or subtree-replace), with
    # detach-attach move being impossible if max_depth == 1
    p_detach_attach = jnp.where(
        kernel_prior.kernel_structure_prior.max_depth == 1,
        jnp.array(0.0),
        p_detach_attach,
    )
    perform_detach_attach = jr.bernoulli(
        detach_attach_key,
        p_detach_attach,
    )

    # perform structure move
    proposed_kernel_state, detach_attach_log_prob = lax.cond(
        perform_detach_attach,
        lambda key: detach_attach_move(
            key,
            kernel_state,
            kernel_prior,
            verbosity=verbosity,
        ),
        lambda key: (
            subtree_replace_move(
                key,
                kernel_state,
                kernel_prior,
                verbosity=verbosity,
            ),
            jnp.array(0.0),  # no detach-attach log prob if subtree-replace,
        ),  # (Preposition 1&2 in Saad2023)
        move_key,
    )

    return proposed_kernel_state, detach_attach_log_prob, perform_detach_attach
